import re
import pprint
import sys
import collections
import copy
import json

import pysam

import importlib

top_package_name = __name__.split(".")[0]
common = importlib.import_module(".".join([top_package_name, "common"]))

infoformat = importlib.import_module(
    ".".join([top_package_name, "variantplus", "infoformat"])
)
varianthandler = importlib.import_module(
    ".".join([top_package_name, "variantplus", "varianthandler"])
)

# annotation = importlib.import_module('.'.join([top_package_name, 'annotation']))
annotitem_misc = importlib.import_module(
    ".".join([top_package_name, "annotation", "misc"])
)
ensembl_parser = importlib.import_module(
    ".".join([top_package_name, "annotation", "ensembl_parser"])
)
ensembl_rest = importlib.import_module(
    ".".join([top_package_name, "annotation", "ensembl_rest"])
)
veplib = importlib.import_module(".".join([top_package_name, "annotation", "veplib"]))
customfile = importlib.import_module(
    ".".join([top_package_name, "annotation", "customfile"])
)
rnalib = importlib.import_module(".".join([top_package_name, "annotation", "rnalib"]))

structvars = importlib.import_module(
    ".".join([top_package_name, "svlib", "structvars"])
)


# CONSTANTS
#SEPS = {
#    "nonprint": {"top": "\x1c", "keyval": "\x1d", "seq": "\x1e", "subkeyval": "\x1f"},
#    "initial": {"top": "$$$", "keyval": ":::", "seq": "&&&", "subkeyval": "@@@"},
#    "percent": {"top": "%3A", "keyval": "%3B", "seq": "%3D", "subkeyval": "%25"},
#}
DEFAULT_SEPTYPE = "nonprint"

# SEP_TOP = '$$$'
# SEP_KEYVAL = ':::'
# SEP_SEQ = '&&&'
# SEP_SUBKEYVAL = '@@@'

# SEP_TOP = '%3A'
# SEP_KEYVAL = '%3B'
# SEP_SEQ = '%3D'
# SEP_SUBKEYVAL = '%25'

# SEP_TOP = '\x1c'
# SEP_KEYVAL = '\x1d'
# SEP_SEQ = '\x1e'
# SEP_SUBKEYVAL = '\x1f'

ATOMIC_TYPES = (type(None), str, int, float, bool)


class AnnotDB:
    """
    Attributes:
        featureDB
        transcript
        transcript_canonical
        regulatory
        motif
        repeat
        popfreq
        cosmic
    """

    def __init__(
        self, annotdb_type, refver, fasta, chromdict, vr=None, septype=DEFAULT_SEPTYPE
    ):
        assert (
            annotdb_type in ANNOTDB_TYPES
        ), f"annotdb_type must be one of: {ANNOTDB_TYPES}"

        self.annotdb_type = annotdb_type
        self.refver = refver
        # assert self.refver in ensembl_rest.REFVERS_ALLOWED
        self.fasta = fasta
        self.chromdict = chromdict
        self.septype = septype

        if vr is None:
            self.init_empty()
        else:
            self.load_vr(vr)

    def __repr__(self):
        return common.cpformat(self.get_showdict())

    def get_showdict(self):
        showdict = dict()
        showdict["popfreq"] = dict(self.popfreq)
        showdict["cosmic"] = dict(self.cosmic)

        showdict["transcript"] = self.transcript.get_showdict()
        showdict["regulatory"] = self.regulatory.get_showdict()
        showdict["motif"] = self.motif.get_showdict()
        showdict["repeat"] = self.repeat.get_showdict()

        return showdict

    def init_empty(self):
        self.transcript = AnnotItemDict(septype=self.septype)
        self.regulatory = AnnotItemDict(septype=self.septype)
        self.motif = AnnotItemDict(septype=self.septype)
        self.repeat = AnnotItemDict(septype=self.septype)

        self.set_transcript_subsets()

        self.popfreq = Popfreq(septype=self.septype)
        self.cosmic = Cosmic(septype=self.septype)

    def load_vr(self, vr):
        self.transcript = AnnotItemDict(
            vr,
            infokey=INFOMETAS[self.annotdb_type]["transcript"]["ID"],
            septype=self.septype,
        )
        self.regulatory = AnnotItemDict(
            vr,
            infokey=INFOMETAS[self.annotdb_type]["regulatory"]["ID"],
            septype=self.septype,
        )
        self.motif = AnnotItemDict(
            vr,
            infokey=INFOMETAS[self.annotdb_type]["motif"]["ID"],
            septype=self.septype,
        )
        self.repeat = AnnotItemDict(
            vr,
            infokey=INFOMETAS[self.annotdb_type]["repeat"]["ID"],
            septype=self.septype,
        )

        self.set_transcript_subsets()

        self.popfreq = Popfreq(
            vr,
            infokey=INFOMETAS[self.annotdb_type]["popfreq"]["ID"],
            septype=self.septype,
        )
        self.cosmic = Cosmic(
            vr,
            infokey=INFOMETAS[self.annotdb_type]["cosmic"]["ID"],
            septype=self.septype,
        )

    def write(self, vr, add_meta=False):
        if add_meta:
            add_infometas(vr.header)

        self.transcript.write_info(
            vr, key=INFOMETAS[self.annotdb_type]["transcript"]["ID"], add_meta=False
        )
        self.regulatory.write_info(
            vr, key=INFOMETAS[self.annotdb_type]["regulatory"]["ID"], add_meta=False
        )
        self.motif.write_info(
            vr, key=INFOMETAS[self.annotdb_type]["motif"]["ID"], add_meta=False
        )
        self.repeat.write_info(
            vr, key=INFOMETAS[self.annotdb_type]["repeat"]["ID"], add_meta=False
        )

        self.popfreq.write_info(
            vr, key=INFOMETAS[self.annotdb_type]["popfreq"]["ID"], add_meta=False
        )
        self.cosmic.write_info(
            vr, key=INFOMETAS[self.annotdb_type]["cosmic"]["ID"], add_meta=False
        )

    # POPFREQ ###################################################

    def update_popfreq(self, vcfspec, dbsnp_vcf, search_equivs=True, overwrite=False):
        assert (
            self.annotdb_type == "plain"
        ), f'This method is available only for annotdb_type "plain".'

        try:
            dbsnp_vr = fetch_dbsnp_vr(vcfspec, dbsnp_vcf, self.fasta, search_equivs)
        except Exception as e:
            raise Exception(f"dbsnp VCF fetch failed: " f"vcfspec: {vcfspec}") from e

        self.popfreq.load_other(parse_dbsnp_vr(dbsnp_vr), overwrite=overwrite)

    # COSMIC #####################################################

    def update_cosmic(
        self,
        vcfspec,
        cosmic_coding_vcf,
        cosmic_noncoding_vcf,
        search_equivs=True,
        overwrite=False,
    ):
        assert (
            self.annotdb_type == "plain"
        ), f'This method is available only for annotdb_type "plain".'

        cosmic_vr, cosmic_vr_noncoding = fetch_cosmic_vr(
            vcfspec,
            cosmic_coding_vcf,
            cosmic_noncoding_vcf,
            self.fasta,
            search_equivs=search_equivs,
        )
        self.cosmic.load_other(
            parse_cosmic_vr(cosmic_vr, cosmic_vr_noncoding), overwrite=overwrite
        )

    # FEATURES ###################################################

    # post-cmdline-vep modifiers

    def update_features_postvep_plain(
        self,
        vcfspec,
        tabixfile_geneset,
        tabixfile_regulatory,
        tabixfile_repeats,
        distance=veplib.DEFAULT_DISTANCE,
    ):
        """To be run with non-SV variants"""
        assert self.annotdb_type == "plain"

        mttype = vcfspec.get_mttype_firstalt()
        assert mttype != "sv", f"Input variant must not be an SV."

        # update using customfiles
        fetch_interval = common.Interval(
            chrom=vcfspec.chrom,
            start1=max(1, vcfspec.pos - distance),
            end1=min(self.chromdict[vcfspec.chrom], vcfspec.pos + distance),
        )
        self.update_customfile_transcript(
            fetch_interval, tabixfile_geneset, fill_missing_is_canonical=True
        )
        self.update_customfile_regulatory(fetch_interval, tabixfile_regulatory)
        self.update_customfile_repeat(fetch_interval, tabixfile_repeats)

        # modifier
        self.modifier_for_plain(vcfspec, mttype)
        self.set_transcript_subsets()

    def update_features_postvep_interval(
        self,
        interval,
        tabixfile_geneset,
        tabixfile_regulatory,
        tabixfile_repeats,
        distance=veplib.DEFAULT_DISTANCE,
    ):
        assert self.annotdb_type == "plain"

        # update using customfiles
        fetch_interval = common.Interval(
            chrom=interval.chrom,
            start1=max(1, interval.start1 - distance),
            end1=min(self.chromdict[interval.chrom], interval.end1 + distance),
        )
        self.update_customfile_transcript(
            fetch_interval, tabixfile_geneset, fill_missing_is_canonical=True
        )
        self.update_customfile_regulatory(fetch_interval, tabixfile_regulatory)
        self.update_customfile_repeat(fetch_interval, tabixfile_repeats)

        # modifier
        self.modifier_for_interval(interval)
        self.set_transcript_subsets()

    def update_features_postvep_bnd(
        self,
        chrom,
        pos,
        endis5,
        tabixfile_geneset,
        tabixfile_regulatory,
        tabixfile_repeats,
        distance=veplib.DEFAULT_DISTANCE,
    ):
        assert self.annotdb_type in ("bnd1", "bnd2")

        # update using customfiles
        if endis5:
            fetch_interval = common.Interval(
                chrom=chrom, start1=pos, end1=min(self.chromdict[chrom], pos + distance)
            )
        else:
            fetch_interval = common.Interval(
                chrom=chrom, start1=max(1, pos - distance), end1=pos
            )

        self.update_customfile_transcript(
            fetch_interval, tabixfile_geneset, fill_missing_is_canonical=True
        )
        self.update_customfile_regulatory(fetch_interval, tabixfile_regulatory)
        self.update_customfile_repeat(fetch_interval, tabixfile_repeats)

        # modifier
        self.modifier_for_bnd(pos, endis5)
        self.set_transcript_subsets()

    # ensembl rest wrappers

    def update_features_ensembl_wrapper_plain(
        self,
        vcfspec,
        distance=veplib.DEFAULT_DISTANCE,
        vep=False,
        overlap=False,
        transcript=False,
        regulatory=False,
    ):
        """To be run with non-SV variants"""

        assert self.annotdb_type == "plain"
        mttype = vcfspec.get_mttype_firstalt()
        assert mttype != "sv", f"Input variant must not be an SV."

        if vep:
            self.update_ensembl_rest_vep_plain(vcfspec, distance)
        if overlap:
            self.update_ensembl_rest_overlap_plain(vcfspec, distance)
        if transcript:
            self.update_ensembl_rest_lookup_transcript()
        if regulatory:
            self.update_ensembl_rest_regulatory()

        self.modifier_for_plain(vcfspec, mttype)
        self.set_transcript_subsets()

    def update_features_ensembl_wrapper_interval(
        self,
        interval,
        distance=veplib.DEFAULT_DISTANCE,
        vep=False,
        overlap=False,
        transcript=False,
        regulatory=False,
    ):
        assert self.annotdb_type == "plain"

        if vep:
            self.update_ensembl_rest_vep_interval(interval, distance)
        if overlap:
            self.update_ensembl_rest_overlap_interval(interval, distance)
        if transcript:
            self.update_ensembl_rest_lookup_transcript()
        if regulatory:
            self.update_ensembl_rest_regulatory()

        self.modifier_for_interval(interval)
        self.set_transcript_subsets()

    def update_features_bnd(
        self,
        chrom,
        pos,
        endis5,
        distance=veplib.DEFAULT_DISTANCE,
        vep=False,
        overlap=False,
        transcript=False,
        regulatory=False,
    ):
        assert self.annotdb_type in ("bnd1", "bnd2")

        if vep:
            self.update_ensembl_rest_vep_bnd(chrom, pos, distance)
        if overlap:
            self.update_ensembl_rest_overlap_bnd(chrom, pos, endis5, distance)
        if transcript:
            self.update_ensembl_rest_lookup_transcript()
        if regulatory:
            self.update_ensembl_rest_regulatory()

        self.modifier_for_bnd(pos, endis5)
        self.set_transcript_subsets()

    # FEATURE UPDATE MODIFIERS ##########################################

    # for setting missing distances

    def set_missing_distances(self, distance_setter):
        """Sets distance for annotitems not derived from VEP (e.g. repeat
        elements). Simulates the distance given by VEP.
        """

        for attrname in ANNOTDB_ATTRNAMES_FEATURES:
            annotitemdict = getattr(self, attrname)
            for key, annotitem in annotitemdict.items():
                if "distance" not in annotitem:
                    # annotitems derived from "ensembl rest overlap"
                    #   do not have "distance" value
                    assert "start0" in annotitem and "end0" in annotitem, (
                        f'"start0" and "end0" values are not set for this '
                        f"AnnotItem object:\n{annotitem}"
                    )
                    feature_range0 = range(annotitem["start0"], annotitem["end0"])
                    annotitem["distance"] = distance_setter(feature_range0)

    def get_distance_setter_ins(self, vcfspec):
        pos0 = vcfspec.pos - 1

        def distance_setter(feature_range0):
            if pos0 <= feature_range0.start - 1:
                distance = feature_range0.start - pos0 - 1
            elif pos0 >= feature_range0.stop - 1:
                distance = feature_range0.stop - 1 - pos0
            else:
                distance = None

            return distance

        return distance_setter

    def get_distance_setter_snv(self, vcfspec):
        pos0 = vcfspec.pos - 1

        def distance_setter(feature_range0):
            if pos0 < feature_range0.start:
                distance = feature_range0.start - pos0
            elif pos0 >= feature_range0.stop:
                distance = feature_range0.stop - 1 - pos0
            else:
                distance = None

            return distance

        return distance_setter

    def get_distance_setter_interval(self, var_range0):
        def distance_setter(feature_range0):
            if var_range0.stop <= feature_range0.start:
                distance = feature_range0.start - var_range0.stop + 1
            elif var_range0.start >= feature_range0.stop:
                distance = feature_range0.stop - 1 - var_range0.start
            else:
                distance = None

            return distance

        return distance_setter

    def get_distance_setter_del(self, vcfspec):
        var_range0 = range(vcfspec.pos, vcfspec.pos + len(vcfspec.ref) - 1)
        return self.get_distance_setter_interval(var_range0)

    def get_distance_setter_bnd(self, pos):
        vcfspec = common.Vcfspec(None, pos, None, None)
        return self.get_distance_setter_snv(vcfspec)

    # modifier for plain
    def modifier_for_plain(self, vcfspec, mttype):
        if mttype == "snv":
            distance_setter = self.get_distance_setter_snv(vcfspec)
        elif mttype == "ins":
            distance_setter = self.get_distance_setter_ins(vcfspec)
        elif mttype in ("del", "delins", "mnv", "cpgmet"):
            distance_setter = self.get_distance_setter_del(vcfspec)
        else:
            raise Exception(f"Unexpected mttype: {mttype}")

        self.set_missing_distances(distance_setter)

    # modifier for interval
    def modifier_for_interval(self, interval):
        distance_setter = self.get_distance_setter_interval(interval.range0)
        self.set_missing_distances(distance_setter)
        self.set_is_enclosed(interval.range0)

    def set_is_enclosed(self, var_range0):
        """
        Sets "is_enclosed" attribute. Must be run after "start0" and "end0"
        values have been set for all annotitems.
        """

        for attrname in ANNOTDB_ATTRNAMES_FEATURES:
            annotitemdict = getattr(self, attrname)
            for key, annotitem in annotitemdict.items():
                feature_range0 = range(annotitem["start0"], annotitem["end0"])
                annotitem["is_enclosed"] = (
                    feature_range0.start >= var_range0.start
                    and feature_range0.stop <= var_range0.stop
                )

    # modifier for breakend
    def modifier_for_bnd(self, pos, endis5):
        distance_setter = self.get_distance_setter_bnd(pos)
        self.set_missing_distances(distance_setter)

        self.settle_border_issues(pos, endis5)

    def settle_border_issues(self, pos, endis5):
        """
        Must be run after "set_missing_distances" and "lookup_transcript" or
        "lookup_regulatory" has been run.

        1) Removes features entirely on the other side
        2) For features whose border coincide with the breakend border,
        distance is changed from None to 0.
        3) Sets "is_broken" attribute
        """

        (on_the_other_side, on_the_border) = self.get_feature_classifiers(endis5)

        for attrname in ANNOTDB_ATTRNAMES_FEATURES:
            annotitemdict = getattr(self, attrname)
            # step 1) and 2)
            keys_to_delete = list()
            for key, annotitem in annotitemdict.items():
                if on_the_other_side(annotitem):
                    keys_to_delete.append(key)
                else:
                    if annotitem["distance"] is None:
                        if on_the_border(annotitem, pos):
                            annotitem["distance"] = 0
            for key in keys_to_delete:
                del annotitemdict[key]
            # step 3)
            for key, annotitem in annotitemdict.items():
                annotitem["is_broken"] = annotitem["distance"] is None

    def get_feature_classifiers(self, endis5):
        """
                (endis5 == True)
                                   THE OTHER SIDE    THE BREAKEND SIDE
                                   - - - - - - - -  -----------------
        FEATURE ON THE OTHER SIDE  = = = = = =     | = = = = = = = =  : FEATURE ON THE BORDER
                                   - - - - - - - -  -----------------
                                   A C G T A C G T   A C G T A C G T
                                                     *(pos)
                ("= = =" denotes a feature)
        """

        if endis5:

            def on_the_other_side(annotitem):
                return (annotitem["distance"] is not None) and (
                    annotitem["distance"] < 0
                )

            def on_the_border(annotitem, pos):
                return annotitem["start1"] == pos

        else:

            def on_the_other_side(annotitem):
                return (annotitem["distance"] is not None) and (
                    annotitem["distance"] > 0
                )

            def on_the_border(annotitem, pos):
                return annotitem["end1"] == pos

        return on_the_other_side, on_the_border

    # CUSTOM FILE FETCH FUNCTIONS #####################################

    def update_customfile_repeat(self, fetch_interval, tabixfile_repeats):
        repeat_dict = customfile.fetch_repeat(
            fetch_interval.chrom,
            fetch_interval.start0,
            fetch_interval.end0,
            tabixfile_repeats,
        )
        self.repeat.load_other(repeat_dict, overwrite=False, create_new=True)

    def update_customfile_transcript(
        self, fetch_interval, tabixfile_geneset, fill_missing_is_canonical=True
    ):
        transcript_dict = customfile.fetch_transcript(
            fetch_interval.chrom,
            fetch_interval.start0,
            fetch_interval.end0,
            tabixfile_geneset,
        )
        self.transcript.load_other(transcript_dict, overwrite=False, create_new=True)
        if fill_missing_is_canonical:
            for annotitem_id, annotitem in self.transcript.items():
                if "is_canonical" not in annotitem:
                    raw_result = ensembl_rest.lookup_id(
                        ID=annotitem["id"], refver=self.refver, expand=False
                    )
                    if raw_result["is_canonical"] == 1:
                        annotitem["is_canonical"] = True
                    else:
                        annotitem["is_canonical"] = False

    def update_customfile_regulatory(self, fetch_interval, tabixfile_regulatory):
        regulatory_dict = customfile.fetch_regulatory(
            fetch_interval.chrom,
            fetch_interval.start0,
            fetch_interval.end0,
            tabixfile_regulatory,
        )
        self.regulatory.load_other(regulatory_dict, overwrite=False, create_new=True)

    # ENSEMBL UNIT FUNCTIONS ##########################################

    # vep
    def update_ensembl_rest_vep_plain(self, vcfspec, distance=veplib.DEFAULT_DISTANCE):
        self._update_ensembl_rest_vep_helper(
            vcfspec=vcfspec, hgvsg=None, distance=distance
        )

    def update_ensembl_rest_vep_interval(
        self, interval, distance=veplib.DEFAULT_DISTANCE
    ):
        sv_del = structvars.Deletion(
            interval.chrom, interval.start1, interval.end1, fasta=self.fasta
        )
        hgvsg = sv_del.get_hgvsg()
        self._update_ensembl_rest_vep_helper(
            vcfspec=None, hgvsg=hgvsg, distance=distance
        )

    def update_ensembl_rest_vep_bnd(self, chrom, pos, distance=veplib.DEFAULT_DISTANCE):
        ref = self.fasta.fetch(chrom, pos - 1, pos)
        vcfspec = common.Vcfspec(chrom, pos, ref, ["N"])
        self._update_ensembl_rest_vep_helper(
            vcfspec=vcfspec, hgvsg=None, distance=distance
        )

    # overlap
    def update_ensembl_rest_overlap_plain(
        self, vcfspec, distance=veplib.DEFAULT_DISTANCE
    ):
        chrom = vcfspec.chrom
        start1 = vcfspec.pos - distance
        end1 = vcfspec.pos + distance
        interval = common.Interval(chrom, start1, end1)
        self._update_ensembl_rest_overlap_helper(interval)

    def update_ensembl_rest_overlap_interval(
        self, interval, distance=veplib.DEFAULT_DISTANCE
    ):
        new_interval = common.Interval(
            chrom=interval.chrom,
            start1=max(1, interval.start1 - distance),
            end1=min(self.chromdict[interval.chrom], interval.end1 + distance),
        )
        self._update_ensembl_rest_overlap_helper(new_interval)

    def update_ensembl_rest_overlap_bnd(
        self, chrom, pos, endis5, distance=veplib.DEFAULT_DISTANCE
    ):
        if endis5:
            start1 = pos
            end1 = min(self.chromdict[chrom], pos + distance)
        else:
            start1 = max(1, pos - distance)
            end1 = pos
        interval = common.Interval(chrom, start1, end1)
        self._update_ensembl_rest_overlap_helper(interval)

    # lookup_transcript
    def update_ensembl_rest_lookup_transcript(self):
        parsed = ensembl_parser.parse_rest_lookup_transcript_post(
            ensembl_rest.lookup_id_post(
                tuple(self.transcript.keys()), refver=self.refver, expand=True
            ),
            refver=self.refver,
            set_gene_name=True,
        )
        self._load_ensembl_parsed(parsed, overwrite=False, create_new=False)

    # regulatory
    def update_ensembl_rest_regulatory(self):
        for ID in self.regulatory.keys():
            self._update_ensembl_rest_regulatory_helper(ID)

    # cmdline VEP
    def update_cmdline_vep(self, vr, overwrite=True, create_new=True):
        try:
            parsed = ensembl_parser.parse_cmdline_vep(vr)
        except Exception as e:
            raise Exception(
                f"Failure of VEP output variant record parsing:" f"\n{vr}"
            ) from e

        self._load_ensembl_parsed(parsed, overwrite=overwrite, create_new=create_new)

    # helpers
    def _load_ensembl_parsed(self, parsed, overwrite=False, create_new=False):
        for annotdb_attrname, other_annotitemdict in parsed.items():
            annotitemdict = getattr(self, annotdb_attrname)
            annotitemdict.load_other(
                other_annotitemdict, overwrite=overwrite, create_new=create_new
            )

    def _update_ensembl_rest_vep_helper(self, vcfspec, hgvsg, distance):
        parsed = ensembl_parser.parse_rest_vep(
            ensembl_rest.vep(
                refver=self.refver,
                vcfspec=vcfspec,
                hgvsg=hgvsg,
                distance=distance,
                with_CADD=True,
                with_Phenotypes=False,
                with_canonical=True,
                with_mane=True,
                with_miRNA=False,
                with_numbers=True,
                with_protein=True,
                with_ccds=True,
                with_hgvs=True,
            )
        )
        self._load_ensembl_parsed(parsed, overwrite=False, create_new=True)

    def _update_ensembl_rest_overlap_helper(self, interval):
        parsed = ensembl_parser.parse_rest_overlap(
            ensembl_rest.overlap(
                chrom=interval.chrom,
                start1=interval.start1,
                end1=interval.end1,
                refver=self.refver,
                transcript=False,
                regulatory=False,
                motif=False,
                repeat=True,
            ),
            refver=self.refver,
            include_motif_without_evidence=False,
        )
        self._load_ensembl_parsed(parsed, overwrite=False, create_new=True)

    def _update_ensembl_rest_lookup_transcript_helper(self, ID):
        parsed = ensembl_parser.parse_rest_lookup_transcript(
            ensembl_rest.lookup_id(ID, refver=self.refver, expand=True),
            refver=self.refver,
            set_gene_name=True,
        )
        self._load_ensembl_parsed(parsed, overwrite=False, create_new=False)

    def _update_ensembl_rest_regulatory_helper(self, ID):
        parsed = ensembl_parser.parse_rest_regulatory(
            ensembl_rest.regulatory(ID, refver=self.refver)
        )
        self._load_ensembl_parsed(parsed, overwrite=False, create_new=False)

    # FEATURE ADDITIONAL ATTRIBUTES GENERATION #####################

    def set_transcript_subsets(self):
        # canonical, overlap, both
        self.transcript_canon = AnnotItemDict()
        self.transcript_ovlp = AnnotItemDict()
        self.transcript_canon_ovlp = AnnotItemDict()

        for ID, annotitem in self.transcript.items():
            if annotitem["is_canonical"]:
                self.transcript_canon[ID] = annotitem
            if annotitem["distance"] is None:
                self.transcript_ovlp[ID] = annotitem
            if annotitem["is_canonical"] and (annotitem["distance"] is None):
                self.transcript_canon_ovlp[ID] = annotitem

    def set_exon_intron_ranges(self):
        def make_exon_ranges(annotitem):
            exon_ranges = dict()
            for key, val in annotitem["exons"].items():
                keysp = key.split("_")
                valsp = [int(x) for x in val.split("_")]
                exon_number = int(keysp[0])
                exon_ranges[exon_number] = range(valsp[0], valsp[1])
            return exon_ranges

        def make_intron_ranges(exon_ranges, annotitem):
            if len(exon_ranges) == 1:
                intron_ranges = None
            else:
                intron_ranges = dict()
                ascending = sorted(exon_ranges.items(), key=lambda x: x[1][0])
                for idx in range(len(ascending) - 1):
                    intron_number = (
                        ascending[idx][0]
                        if annotitem["is_forward"]
                        else ascending[idx + 1][0]
                    )
                    range_item = range(
                        ascending[idx][1][-1] + 1, ascending[idx + 1][1][0]
                    )
                    intron_ranges[intron_number] = range_item
                intron_ranges = dict(sorted(intron_ranges.items(), key=lambda x: x[0]))

            return intron_ranges

        for ID, annotitem in self.transcript.items():
            if "exons" in annotitem.keys():
                annotitem.exon_ranges = make_exon_ranges(annotitem)
                annotitem.intron_ranges = make_intron_ranges(
                    annotitem.exon_ranges, annotitem
                )


class AnnotItemBase:
    def __repr__(self):
        return common.cpformat(self.get_self_show(), sort_dicts=False)

    @property
    def annotkey(self):
        return self.__class__.meta["ID"]

    @classmethod
    def get_annotkey(cls):
        return cls.meta["ID"]

    @classmethod
    def add_meta_info(cls, vcfheader):
        vcfheader.add_meta(key="INFO", items=cls.meta.items())

    @classmethod
    def add_meta_format(cls, vcfheader):
        vcfheader.add_meta(key="FORMAT", items=cls.meta.items())

    def write_info(self, vr, key=None, add_meta=False):
        if key is None:
            key = self.annotkey
        if add_meta:
            self.__class__.add_meta_info(vr.header)

        infoformat.set_info(vr, key, self.encode())

    def write_format(self, vr, sampleid, key=None, add_meta=False):
        if key is None:
            key = self.annotkey
        if add_meta:
            self.__class__.add_meta_format(vr.header)

        infoformat.set_format(vr, sampleid, key, self.encode())

    def write_header(self, vcfheader):
        vcfheader.add_meta(key=self.annotkey, value=self.encode())


class AnnotItem(AnnotItemBase, dict):
    meta = None
    # to be used as "items" argument for pysam.VariantHeader.add_meta
    SEPS = {
        "nonprint": {"top": "\x1c", "keyval": "\x1d", "seq": "\x1e", "subkeyval": "\x1f"},
        "prototype": {"top": "$$$", "keyval": ":::", "seq": "&&&", "subkeyval": "@@@"},
        "percent": {"top": "%3A", "keyval": "%3B", "seq": "%3D", "subkeyval": "%25"},
    }

    def __init__(self, vr=None, infokey=None, septype=DEFAULT_SEPTYPE, **kwargs):
        # super().__init__()
        self.septype = septype
        # if (vr is not None):
        #    self.load_vr(vr, infokey)

    def __setitem__(self, key, val):
        # self._sanitycheck_keyval(key, val)
        super().__setitem__(key, val)

    def __getitem__(self, key):
        return super().__getitem__(key)

    #infokey = annotkey  # alias

    ##### constructors #####
    @classmethod
    def decode(cls, infostring, septype=None):
        if septype is None:
            septype = DEFAULT_SEPTYPE

        sep_top = cls.SEPS[septype]["top"]
        sep_keyval = cls.SEPS[septype]["keyval"]
        sep_seq = cls.SEPS[septype]["seq"]
        sep_subkeyval = cls.SEPS[septype]["subkeyval"]

        annotitem = cls()
        for x in infostring.split(sep_top):
            x_sp = x.split(sep_keyval)
            annotitem[x_sp[0]] = x_sp[1]
        for key, val in annotitem.items():
            if len(val.split(sep_subkeyval)) == 1:  # not a dict
                new_val = [common.str_to_nonstr(x) for x in val.split(sep_seq)]
                if len(new_val) == 1:
                    new_val = new_val[0]
                else:
                    new_val = new_val[1:-1]
            else:
                tmp = list()
                for x in val.split(sep_seq):
                    x_split = x.split(sep_subkeyval)
                    tmp_item = map(common.str_to_nonstr, x_split)
                    tmp.append(tmp_item)
                new_val = dict(tmp)

            key = common.str_to_nonstr(key)

            annotitem[key] = new_val

        return annotitem

    @classmethod
    def from_vcfheader(cls, vcfheader):
        result = cls()
        annotkey = result.get_annotkey()
        for rec in vcfheader.records:
            if rec.key == annotkey:
                result.load_infostring(rec.value)

        return result
    ########################

    def encode(self, septype=None):
        if septype is None:
            septype = self.septype

        sep_top = self.__class__.SEPS[septype]["top"]
        sep_keyval = self.__class__.SEPS[septype]["keyval"]
        sep_seq = self.__class__.SEPS[septype]["seq"]
        sep_subkeyval = self.__class__.SEPS[septype]["subkeyval"]

        if len(self) == 0:
            infostring = None
        else:
            result = list()
            for key, val in self.items():
                self._sanitycheck_keyval(key, val)
                if isinstance(val, (tuple, list)):
                    modified_val = sep_seq + sep_seq.join(map(str, val)) + sep_seq
                elif isinstance(val, dict):
                    modified_val = sep_seq.join(
                        sep_subkeyval.join(map(str, x)) for x in val.items()
                    )
                else:
                    modified_val = val

                result.append(f"{key}{sep_keyval}{modified_val}")

            infostring = sep_top.join(result)

        return infostring

    def _sanitycheck_comma_semicolon(self, x):
        if isinstance(x, str):
            assert ("," not in x) and (
                ";" not in x
            ), f'AnnotItem components must not contain "," or ";".'

    def _sanitycheck_keyval(self, key, val):
        try:
            assert isinstance(
                key, ATOMIC_TYPES
            ), f"The key must be one of {ATOMIC_TYPES}."
            self._sanitycheck_comma_semicolon(key)

            if isinstance(val, dict):
                for subkey, subval in val.items():
                    assert isinstance(subkey, ATOMIC_TYPES), (
                        f"When the value is a dict, each subkey must be "
                        f"one of {ATOMIC_TYPES}"
                    )
                    self._sanitycheck_comma_semicolon(subkey)

                    assert isinstance(subval, ATOMIC_TYPES), (
                        f"When the value is a dict, the type of each "
                        f"subvalue must be one of {ATOMIC_TYPES}."
                    )
                    self._sanitycheck_comma_semicolon(subval)
            elif isinstance(val, list):
                for subval in val:
                    assert isinstance(subval, ATOMIC_TYPES), (
                        f"When the value is a list, the type of each item "
                        f"must be one of {ATOMIC_TYPES}."
                    )
                    self._sanitycheck_comma_semicolon(subval)
            else:
                assert isinstance(val, ATOMIC_TYPES), (
                    f"When the value is neither list nor dict, its type "
                    f"must be one of {ATOMIC_TYPES}."
                )
                self._sanitycheck_comma_semicolon(val)

        except AssertionError as e:
            new_e = Exception(
                f"Invalid key/value for assignment to "
                f"AnnotItem: key - {key} ; value - {val}"
            )
            raise new_e from e

    def load_infostring(self, infostring, septype=None, overwrite=True):
        other_annotitem = self.decode(infostring, septype=septype)
        self.load_other(other_annotitem, overwrite=overwrite)

    def load_vr_info(self, vr, infokey=None):
        if infokey is None:
            infokey = self.annotkey

        assert vr.header.info[infokey].number == 1
        assert vr.header.info[infokey].type == "String"

        if not infoformat.check_NA_info(vr, infokey):
            infostring = infoformat.get_info(vr, infokey)
            self.load_infostring(infostring)

    load_vr = load_vr_info  # alias

    def load_vr_format(self, vr, sampleid, key=None):
        if key is None:
            key = self.annotkey

        assert vr.header.formats[key].number == 1
        assert vr.header.formats[key].type == "String"

        if not infoformat.check_NA_format(vr, sampleid, key):
            infostring = infoformat.get_format(vr, sampleid, key)
            self.load_infostring(infostring)

    def load_other(self, other, overwrite=False):
        """Args:
        other: an AnnotItem object
        """
        if overwrite:
            for key, val in other.items():
                self[key] = val
        else:
            for key, val in other.items():
                try:
                    old_val = self[key]
                except KeyError:
                    self[key] = val
                else:
                    if old_val is None:
                        self[key] = val
                    else:
                        if val != old_val:
                            raise Exception(
                                f"Old value and new value are different; "
                                f"key: {key} ; old_val: {old_val} ; "
                                f"new_val: {val}"
                            )

#    def write_info(self, vr, key=None, add_meta=False):
#        if key is None:
#            key = self.annotkey
#        if add_meta:
#            self.__class__.add_meta_info(vr.header)
#
#        infoformat.set_info(vr, key, self.encode())
#
#    # write = write_info  # alias
#
#    def write_format(self, vr, sampleid, key=None, add_meta=False):
#        if key is None:
#            key = self.annotkey
#        if add_meta:
#            self.__class__.add_meta_format(vr.header)
#
#        infoformat.set_format(vr, sampleid, key, self.encode())


class AnnotItemList(AnnotItemBase, list):
    meta = None

    def get_self_show(self):
        return [(None if (x is None) else x.get_self_show())
                for x in self]

    def encode(self):
        return tuple((None if (x is None) else x.encode())
                     for x in self)


class AnnotItemDict(dict):
    def __init__(self, vr=None, infokey=None, septype=DEFAULT_SEPTYPE):
        # super().__init__()
        self.septype = septype
        if (vr is not None) and (infokey is not None):
            self.load_vr(vr, infokey)

    def __setitem__(self, key, val):
        assert isinstance(val, AnnotItem)
        super().__setitem__(key, val)

    def __repr__(self):
        return common.cpformat(self.get_showdict())

    def get_showdict(self):
        showdict = {key: dict(val) for (key, val) in self.items()}
        return showdict

    def get_annotitem_id(self, annotitem):
        pass

    def load_vr(self, vr, infokey):
        assert vr.header.info[infokey].number == "."
        assert vr.header.info[infokey].type == "String"

        if not infoformat.check_NA_info(vr, infokey):
            infoval = infoformat.get_info(vr, infokey, collapse_tuple=False)
            for infostring in infoval:
                annotitem = AnnotItem(septype=self.septype)
                annotitem.load_infostring(infostring)
                ID = self.get_annotitem_id(annotitem)
                self[ID] = annotitem

    def load_other(self, other, overwrite=True, create_new=False):
        for ID, annotitem in other.items():
            if ID in self.keys():
                self[ID].load_other(annotitem, overwrite=overwrite)
            else:
                if create_new:
                    self[ID] = annotitem

    def write_info(self, vr, key, add_meta=False):
        if add_meta:
            add_infometas(vr.header)

        infoval = list()
        for annotitem in self.values():
            infoval.append(annotitem.encode())

        vr.info[key] = infoval



import importlib
top_package_name = __name__.split('.')[0]
common = importlib.import_module('.'.join([top_package_name, 'common']))
readhandler = importlib.import_module('.'.join([top_package_name, 'readplus', 'readhandler']))


# level 1 functions

def get_pos0_flanking_variant(POS0, REF, ALT, mttype = None):
	'''
	- returns 0-based positions just before/after the bases subject to variation.
	- e.g. if REF == 'ACT' and ALT == 'A', 
	  pos0_before_var is the position of 'A',
	  and pos0_after_var is the position after 'T'.
	- e.g. if REF == 'A' and ALT == 'G', 
	  pos0_before_var is the position before 'A',
	  and pos0_after_var is the position after 'A'.
	- flank length == 1
	'''

	if mttype is None:
		mttype = common.get_mttype(REF, ALT)

	if mttype == 'sv':
		pos0_before_var = None
		pos0_after_var = None
	else:
		if REF[0] == ALT[0]:
			pos0_before_var = POS0
		else:
			pos0_before_var = POS0 - 1

		pos0_after_var = POS0 + len(REF)

	return pos0_before_var, pos0_after_var


# deprecated version #

#def get_pos0_flanking_variant(POS0, REF, ALT, mttype = None):
#	'''
#	- returns 0-based positions just before/after the bases subject to variation.
#	- e.g. if REF == 'ACT' and ALT == 'A', 
#	  pos0_before_var is the position of 'A',
#	  and pos0_after_var is the position after 'T'.
#	- e.g. if REF == 'A' and ALT == 'G', 
#	  pos0_before_var is the position before 'A',
#	  and pos0_after_var is the position after 'A'.
#	- flank length == 1
#	'''
#
#	if mttype == None:
#		mttype = common_funcs.get_mttype(REF, ALT)
#
#	if mttype == 'SV':
#		pos0_before_var = None
#		pos0_after_var = None
#	else:
#		pos0_after_var = POS0 + len(REF)
#	
#		if mttype == 'SNV' or mttype == 'MNV':
#			pos0_before_var = POS0 - 1
#		elif mttype == 'INS' or mttype == 'DEL':
#			pos0_before_var = POS0
#		else: # CINDEL
#			if REF[0] == ALT[0]: 
#				'''
#				Not sure about this but CINDEL REF and ALT may have identical first letter. 
#				In this case, the first position in REF would not be a position
#				undergoing variation.
#				'''
#				pos0_before_var = POS0
#			else:
#				pos0_before_var = POS0 - 1
#
#	return pos0_before_var, pos0_after_var

######################


def generate_pos0_flanking_variant_if_missing(pos0_before_var, pos0_after_var, POS0, REF, ALT):
	if pos0_before_var == None or pos0_after_var == None:
		pos0_before_var, pos0_after_var = get_pos0_flanking_variant(POS0, REF, ALT)
	return pos0_before_var, pos0_after_var


def get_pos0_REFbases(POS0, REF):
	pos0_first_REFbase = POS0
	pos0_last_REFbase = POS0 + len(REF) - 1
	return pos0_first_REFbase, pos0_last_REFbase


def check_read_spans(pairs_dict, pos0):
	return pos0 in pairs_dict['refpos0']


def get_pairs_index(pairs_dict, pos0):
	if check_read_spans(pairs_dict, pos0):
		return pairs_dict['refpos0'].index(pos0)
	else:
		return None


def get_querypos0(pairs_dict, pos0):
	if check_read_spans(pairs_dict, pos0):
		pairs_index = get_pairs_index(pairs_dict, pos0)
		return pairs_dict['querypos0'][pairs_index]
	else:
		return None


def check_read_matches(pairs_dict, pos0):
	# returns True if matches
	# returns False if mismatches or does not span
	if check_read_spans(pairs_dict, pos0):
		pairs_index = get_pairs_index(pairs_dict, pos0)
		return pairs_dict['refseq'][pairs_index].isupper()
	else:
		return False


def check_chrom(read, CHROM):
	return read.reference_name == CHROM


# level 2 functions: utilizes level 1 functions

def check_read_spans_REF_positions(
		read, pairs_dict, CHROM, POS0, REF,
		pos0_first_REFbase = None, 
		pos0_last_REFbase = None,
		):
	if not check_chrom(read, CHROM):
		return False
	if pos0_first_REFbase == None or pos0_last_REFbase == None:
		pos0_first_REFbase, pos0_last_REFbase = get_pos0_REFbases(POS0, REF)

	if \
	check_read_spans(pairs_dict, pos0_first_REFbase) and \
	check_read_spans(pairs_dict, pos0_last_REFbase):
		return True
	else:
		return False


def check_read_partially_spans_REF_positions(
		read, pairs_dict, CHROM, POS0, REF,
		pos0_first_REFbase = None, 
		pos0_last_REFbase = None,
		):
	if not check_chrom(read, CHROM):
		return False
	if pos0_first_REFbase == None or pos0_last_REFbase == None:
		pos0_first_REFbase, pos0_last_REFbase = get_pos0_REFbases(POS0, REF)

	if check_read_spans_REF_positions(
		read, pairs_dict, CHROM, POS0, REF,
		pos0_first_REFbase, pos0_last_REFbase,
		):
		return False
	else:
		for pos0_candidate in range(pos0_first_REFbase, pos0_last_REFbase + 1):
			if check_read_spans(pairs_dict, pos0_candidate):
				return True
		return False


def check_read_flanks_variant(
		read, pairs_dict, CHROM, POS0, REF, ALT,
		pos0_before_var = None, 
		pos0_after_var = None,
		):
	if not check_chrom(read, CHROM):
		return False
	pos0_before_var, pos0_after_var = \
			generate_pos0_flanking_variant_if_missing(pos0_before_var, pos0_after_var, POS0, REF, ALT)

	if \
	check_read_spans(pairs_dict, pos0_before_var) and \
	check_read_spans(pairs_dict, pos0_after_var):
		return True
	else:
		return False


def check_read_matchflanks_variant(
		read, pairs_dict, CHROM, POS0, REF, ALT,
		pos0_before_var = None, 
		pos0_after_var = None,
		):
	if not check_chrom(read, CHROM):
		return False
	pos0_before_var, pos0_after_var = \
			generate_pos0_flanking_variant_if_missing(pos0_before_var, pos0_after_var, POS0, REF, ALT)

	if \
	check_read_matches(pairs_dict, pos0_before_var) and \
	check_read_matches(pairs_dict, pos0_after_var):
		return True
	else:
		return False


def check_read_doesnot_give_information(
		read, CHROM, POS0, REF, ALT,
		pos0_before_var = None, 
		pos0_after_var = None,
		):
	if not check_chrom(read, CHROM):
		return True
	pos0_before_var, pos0_after_var = \
			generate_pos0_flanking_variant_if_missing(pos0_before_var, pos0_after_var, POS0, REF, ALT)

	if \
	read.reference_end - 1 <= pos0_before_var or \
	read.reference_start >= pos0_after_var:
		return True
	else:
		return False


def check_read_gives_ambiguous_information(
		read, pairs_dict, CHROM, POS0, REF, ALT,
		pos0_before_var = None, 
		pos0_after_var = None,
		):
	if not check_chrom(read, CHROM):
		return False
	pos0_before_var, pos0_after_var = \
			generate_pos0_flanking_variant_if_missing(pos0_before_var, pos0_after_var, POS0, REF, ALT)

	if \
	( not check_read_flanks_variant(
		read, pairs_dict, CHROM, POS0, REF, ALT, 
		pos0_before_var = pos0_before_var,
		pos0_after_var = pos0_after_var,
		) ) \
	and \
	( not check_read_doesnot_give_information(
		read, CHROM, POS0, REF, ALT, 
		pos0_before_var = pos0_before_var,
		pos0_after_var = pos0_after_var,
		) ):
		return True
	else:
		return False


# level 3 functions: utilizes level 1 and level 2 functions


# The function below is deprecated.
# It is a version of get_seq_REFrange which returns a valid value
# even if the read does not flank the varying base positions.

#def get_seq_REFrange(read, CHROM, POS0, REF, ALT, pairs_dict = None):
#	'''
#	- Does not require the read to flank the positions of the varying bases,
#	  but the read must span over all REF positions to get a valid result.
#	- If chromosomes do not match, result is None.
#	- If the read does not span over all REF positions, result is None.
#	- Else, if all REF positions are cigar D, result is an empty string.
#	'''
#	if pairs_dict == None:
#		pairs_dict = readhandler.get_pairs_dict(read)
#
#	if not check_read_spans_REF_positions(read, pairs_dict, CHROM, POS0, REF):
#		seq_REFrange = None
#	else:
#		# get sliceidx_1
#		pos0_first_REFbase, pos0_last_REFbase = get_pos0_REFbases(POS0, REF)
#		sliceidx_1 = get_pairs_index(pairs_dict, pos0_first_REFbase)
#
#		# get sliceidx_2
#		if read.reference_end == pos0_last_REFbase + 1: # no downstream flanking
#			sliceidx_2 = get_pairs_index(pairs_dict, pos0_last_REFbase) + 1
#		else:
#			pos0_before_var, pos0_after_var = get_pos0_flanking_variant(POS0, REF, ALT)
#			sliceidx_2 = get_pairs_index(pairs_dict, pos0_after_var)
#
#		querypos0_list = [ 
#				x 
#				for x in pairs_dict['querypos0'][sliceidx_1:sliceidx_2]
#				if x != None
#				]
#
#		if len(querypos0_list) == 0: # all REF positions are cigar D (deletion)
#			seq_REFrange = ''
#		else:
#			querypos0_first = querypos0_list[0]
#			querypos0_last = querypos0_list[-1]
#			seq_REFrange = read.query_sequence[ querypos0_first : querypos0_last + 1 ]
#
#	return seq_REFrange


######################################################################################
# both seq_REFrange and alleleclass are None iff the read does not flank the variant #
######################################################################################

def get_querypos0_REFrange_list_from_read(
		POS0, REF, ALT, 
		pairs_dict = None,
		pos0_before_var = None, 
		pos0_after_var = None,
		):
	'''
	Returns a range object representing indices of read.query_sequence,
	which by slicing results in seq_REFrange.
	'''
	pos0_before_var, pos0_after_var = \
			generate_pos0_flanking_variant_if_missing(pos0_before_var, pos0_after_var, POS0, REF, ALT)

	sliceidx_1 = get_pairs_index(pairs_dict, POS0)
	sliceidx_2 = get_pairs_index(pairs_dict, pos0_after_var)

	querypos0_REFrange_list = [ 
			x 
			for x in pairs_dict['querypos0'][sliceidx_1:sliceidx_2]
			if x != None
			]

	return querypos0_REFrange_list


def get_seq_REFrange_from_read(
		read, CHROM, POS0, REF, ALT, 
		pairs_dict = None,
		pos0_before_var = None, 
		pos0_after_var = None,
		):
	'''
	- If the read does not flank the varying base positions, result is None.
	- The flanking positions need not be cigar M.
	- If chromosomes do not match, result is None.
	- If cigar operations on all REF positions are D, result is an empty string.
	'''
	if pairs_dict == None:
		pairs_dict = readhandler.get_pairs_dict(read)
		if pairs_dict == None:
			return None

	pos0_before_var, pos0_after_var = \
			generate_pos0_flanking_variant_if_missing(pos0_before_var, pos0_after_var, POS0, REF, ALT)

	if not check_read_flanks_variant(
			read, pairs_dict, CHROM, POS0, REF, ALT,
			pos0_before_var = pos0_before_var,
			pos0_after_var = pos0_after_var,
			):
		seq_REFrange = None
	else:
		#pos0_first_REFbase, pos0_last_REFbase = get_pos0_REFbases(POS0, REF)
		sliceidx_1 = get_pairs_index(pairs_dict, POS0)
		sliceidx_2 = get_pairs_index(pairs_dict, pos0_after_var)

		querypos0_list = [ 
				x 
				for x in pairs_dict['querypos0'][sliceidx_1:sliceidx_2]
				if x != None
				]

		if len(querypos0_list) == 0: # all REF positions are cigar D (deletion)
			seq_REFrange = ''
		else:
			seq_REFrange = read.query_sequence[ querypos0_list[0] : querypos0_list[-1] + 1 ]

	return seq_REFrange


def get_alleleclass_from_read(
		read, CHROM, POS0, REF, ALT, 
		pairs_dict = None, 
		pos0_before_var = None, 
		pos0_after_var = None,
		seq_REFrange = None,
		):
	'''
	None : read is not qualified to support any allele
	0 : supports REF
	1 : supports ALT
	2 : supports another allele different from REF or ALT
	'''
	# exclude SV
	mttype = common.get_mttype(REF, ALT)
	if mttype == 'sv':
		alleleclass = None
		return alleleclass

	# get pairs_dict if absent
	if pairs_dict == None:
		pairs_dict = readhandler.get_pairs_dict(read)
		if pairs_dict == None:
			return None

	# get pos0 before/after var if absent
	pos0_before_var, pos0_after_var = \
			generate_pos0_flanking_variant_if_missing(pos0_before_var, pos0_after_var, POS0, REF, ALT)

	# read which does not flank the variant is regarded as not qualified to support any allele (flank length == 1)
	if not check_read_flanks_variant(
			read, pairs_dict, CHROM, POS0, REF, ALT,
			pos0_before_var = pos0_before_var,
			pos0_after_var = pos0_after_var,
			):
		alleleclass = None
	else:
		# if the flanking bases are not cigar M, the read is regarded not to support REF nor ALT
		if not check_read_matchflanks_variant(
				read, pairs_dict, CHROM, POS0, REF, ALT,
				pos0_before_var = pos0_before_var,
				pos0_after_var = pos0_after_var,
				):
			alleleclass = 2
		else:
			# evaluate seq_REFrange if not given as argument
			if seq_REFrange == None:
				seq_REFrange = get_seq_REFrange_from_read(
						read, CHROM, POS0, REF, ALT, 
						pairs_dict = pairs_dict,
						pos0_before_var = pos0_before_var,
						pos0_after_var = pos0_after_var,
						)
			if seq_REFrange == None: # this should not happen
				alleleclass = None
				#raise Exception('seq_REFrange is None after check_read_matchflanks_variant function returned True')
			else:
				if seq_REFrange == REF:
					alleleclass = 0
				elif seq_REFrange == ALT:
					alleleclass = 1
				else:
					alleleclass = 2

	return alleleclass




import re

import pysam

import importlib
top_package_name = __name__.split('.')[0]
customfile = importlib.import_module('.'.join([top_package_name, 'annotation', 'customfile']))


def get_next_codon_frame(chrom, pos, endis5, transcript_id, is_forward, tabixfile_geneset):
	next_CDS = get_next_CDS(chrom, pos, transcript_id, tabixfile_geneset, endis5)
	if endis5:
		return get_leftmost_codon_frame(next_CDS, is_forward)
	else:
		return get_rightmost_codon_frame(next_CDS, is_forward)


def get_rna_range(chrom, pos, transcript_id, tabixfile_geneset):
	line = customfile.fetch_transcript_tabixline(chrom, pos - 1, pos, 
												 [transcript_id], 
												 tabixfile_geneset)[0]

	return range(line.start, line.end)


def get_next_CDS(chrom, pos, transcript_id, tabixfile_geneset, endis5):
	"""
	Returns:
		A pysam.libctabixproxies.GTFProxy object representing 
		None if there is no nearby CDS.
	"""

	rna_range = get_rna_range(chrom, pos, transcript_id, tabixfile_geneset)

	if endis5:
		def filter_other_side(line, pos):
			return line.start > (pos - 1)
	else:
		def filter_other_side(line, pos):
			return (line.end - 1) < (pos - 1)
											
	lines_other_side = list()
	for line in tabixfile_geneset.fetch(chrom, rna_range.start, rna_range.stop):
		if line.feature == 'CDS' and re.search(f'Parent=transcript:{transcript_id}', line.attributes):
			if filter_other_side(line, pos):
				lines_other_side.append(line)

	protein_ids = set(customfile.parse_gff3_attrs(x.attributes)['protein_id'] for x in 
					  lines_other_side)
	if len(protein_ids) > 1:
		raise Exception(f'More than one protein IDs for transcript ID "{transcript_id}"')

	if len(lines_other_side) == 0:
		return None
	else:
		if endis5:
			return min(lines_other_side, key = lambda x: x.start)
		else:
			return max(lines_other_side, key = lambda x: x.start)


def phase_to_frame0(phase): 
	"""
	phase: As per ensembl gff3 spec. (http://asia.ensembl.org/info/website/upload/gff3.html)
		If the feature is on the reverse strand, when phase is 1, the base on the left of the rightmost base is the first base of a codon.
	frame: the first base of a codon is 0, the second 1, the third 2
	"""

	return -phase % 3


def get_leftmost_codon_frame0(CDS):
	if CDS.strand == '+':
		return phase_to_frame0(CDS.frame)
	else:
		return (phase_to_frame0(CDS.frame) + (CDS.end - CDS.start - 1))%3


def get_rightmost_codon_frame0(CDS):
	if CDS.strand == '+':
		return (phase_to_frame0(CDS.frame) + (CDS.end - CDS.start - 1))%3
	else:
		return phase_to_frame0(CDS.frame)


##################################

# for bnd annotation

def get_next_codon_frame0(transcript_id_list, chrom, pos, endis5, tabixfile_geneset):
	def get_fetch_range(tabixfile_geneset, chrom, pos):
		range_list = list()
		for tabixline in tabixfile_geneset.fetch(chrom, pos - 1, pos):
			if customfile.check_tabixline_type(tabixline, 'gene'):
				range_list.append(range(tabixline.start, tabixline.end))

		return range(min(x.start for x in range_list),
					 max(x.stop for x in range_list))

	def classify_fetched(tabixfile_geneset, transcript_id_list, chrom, fetch_range):
		fetched_dict = dict()
		for tabixline in tabixfile_geneset.fetch(chrom, fetch_range.start, fetch_range.stop):
			parent = customfile.get_parent(tabixline)
			if parent in transcript_id_list:
				fetched_dict.setdefault(parent, list())
				fetched_dict[parent].append(tabixline)

		return fetched_dict

	def get_next_CDS(pos, transcript_id, fetched_dict, endis5):
		if endis5:
			relevant_features = [x for x in fetched_dict[transcript_id]
								 if x.start + 1 > pos]
			next_pos = min(x.start for x in relevant_features) + 1
			next_features = [x for x in relevant_features if x.start == next_pos - 1]
		else:
			relevant_features = [x for x in fetched_dict[transcript_id]
								 if x.end < pos]
			next_pos = max(x.end for x in relevant_features)
			next_features = [x for x in relevant_features if x.end == next_pos]

		next_feature_roles = [x.feature for x in next_features]
		if next_feature_roles.count('exon') != 1:
			raise Exception(f'Unexpected  exon is not included among the next features:\ntranscript_id: {transcript_id}, pos: {pos}, endis5: {endis5}\nnext_features: {next_features}')
			
		if 'exon' not in next_feature_roles:
			raise Exception(f'exon is not included among the next features:\ntranscript_id: {transcript_id}, pos: {pos}, endis5: {endis5}\nnext_features: {next_features}')

		next_CDS_candidates = [x for x in next_features if x.feature == 'CDS']

		if len(next_CDS_candidates) == 0:
			return None
		elif len(next_CDS_candidates) == 1:
			return next_CDS_candidates[0]
		else:
			raise Exception(f'More than one next CDSs with identical positions:\ntranscript_id="{transcript_id}", pos="{pos}", endis5="{endis5}"')

	def get_result(transcript_id_list, pos, fetched_dict, endis5):
		if endis5:
			frame0_getter = get_leftmost_codon_frame0
		else:
			frame0_getter = get_rightmost_codon_frame0

		result = dict()

		for transcript_id in transcript_id_list:
			next_CDS = get_next_CDS(pos, transcript_id, fetched_dict, endis5)
			if next_CDS is None:
				next_codon_frame0 = None
			else:
				next_codon_frame0 = frame0_getter(next_CDS)
			result[transcript_id] = next_codon_frame0

		return result

	# main
	fetch_range = get_fetch_range(tabixfile_geneset, chrom, pos)
	fetched_dict = classify_fetched(tabixfile_geneset, transcript_id_list, chrom, fetch_range)

	return get_result(transcript_id_list, pos, fetched_dict, endis5)




from math import nan
import os
import numpy as np
import pandas as pd
import warnings
from scipy import signal as scisig
from numpy import matlib
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from scipy.stats import binom
from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.patches as mplpatches
import matplotlib.ticker as ticker
from matplotlib.ticker import LinearLocator
import matplotlib.lines as lines
import matplotlib.transforms as transforms
import string
import warnings
#import seaborn as sns
import sys
pd.options.mode.chained_assignment = None
warnings.filterwarnings("ignore")

def unique_py(seqlist):
    seen = set()
    seen_add = seen.add
    return [x for x in seqlist if not (x in seen or seen_add(x))]

##COMPUTATION OF INTERMUTATIONAL DISTANCE
#check computation of IMD
#major difference is whether its the closest breakpoint or the breakpoint immediately preceding it

def calcIntermutDist2(subs_type,first_chrom_na=False):
    subs_type_processed = subs_type.copy()
    chr_list = unique_py(subs_type['chr'])
    pos_array_im = subs_type['position'].values
    index_orig_df = np.arange(len(subs_type_processed))
    #args_pos_list = np.argsort(pos_array_im)
    args_pos_list=[]
    distPrev_list =[]
    prevPos_list =[]

    for c in chr_list:
        inds_chr = np.where(subs_type['chr']==c)
        pos_array_im_c = np.sort(pos_array_im[inds_chr])
        index_orig_df[inds_chr] = index_orig_df[inds_chr][np.argsort(pos_array_im[inds_chr])]

        if first_chrom_na:
            prevPos_arr_c =  np.hstack((np.NAN,pos_array_im_c.flatten()[:-1]))
        else:
            prevPos_arr_c = np.hstack((0,pos_array_im_c.flatten()[:-1]))
        distPrev_arr_c = pos_array_im_c  - prevPos_arr_c
        distPrev_arr_c[distPrev_arr_c==0] = 1
        distPrev_list = np.append(distPrev_list,distPrev_arr_c.astype(int)).flatten()
        prevPos_list = np.append(prevPos_list,prevPos_arr_c.astype(int)).flatten()
        prevPos_arr_c = []
        distPrev_arr_c = []
    subs_type_processed = subs_type_processed.reindex(index_orig_df).reset_index(drop=True)
    subs_type_processed['prevPos'] = prevPos_list
    subs_type_processed['distPrev'] = distPrev_list
    return subs_type_processed

def calcIntermutDist(subs_type,first_chrom_na=False):
    subs_type_processed = pd.DataFrame()
    for c in unique_py(subs_type['chr']):
        subs_type_chrom = subs_type[subs_type['chr']==c].sort_values('position')
        if first_chrom_na:
            subs_type_chrom['prevPos'] = np.hstack((np.NAN,subs_type_chrom['position'].values.flatten()[:-1]))
        else:
            subs_type_chrom['prevPos'] = np.hstack((0,subs_type_chrom['position'].values.flatten()[:-1]))
        subs_type_chrom['distPrev'] = subs_type_chrom['position'].values - subs_type_chrom['prevPos'].values
        subs_type_processed = subs_type_processed.append(subs_type_chrom)
        subs_type_processed['distPrev'][subs_type_processed['distPrev']==0] = 1
    return subs_type_processed

def computeIMD2(chrom_df, chromosome):

    #keep track of partners

    d1 = dict(zip(list(chrom_df['start1']), list(chrom_df['start2'])))
    d2 = dict(zip(list(chrom_df['start2']), list(chrom_df['start1'])))
    #d = {**d1, **d2} #combine dictionaries, THIS ONLY WORKS IN PYTHON 3.5+
    d = d1.copy()
    d.update(d2)

    lb = chrom_df.iloc[:, 0:2] #get chrom1 and start1
    rb = chrom_df.iloc[:, 3:5] #get chrom2 and start2
    rest = chrom_df.iloc[:, 6:]

    lb = pd.DataFrame(np.concatenate((lb.values, rest.values), axis=1))
    rb = pd.DataFrame(np.concatenate((rb.values, rest.values), axis=1))

    #BREAKPOINTS ARE CONSIDERED INDIVIDUALLY

    #['chrom1', 'start1', 'end1', 'chrom2', 'start2', 'end2', 'sample', 'svclass', 'size_bin', 'length']
    lb.columns = ['chrom1', 'start1', 'sample', 'svclass', 'size_bin', "length"]
    rb.columns = ['chrom2', 'start2', 'sample', 'svclass', 'size_bin', "length"]

    chr_lb = lb[lb.chrom1 == chromosome]
    chr_rb = rb[rb.chrom2 == chromosome]
    # print(chr_lb)
    # print(chr_rb)
    chrom_df = pd.DataFrame(np.concatenate((chr_lb.values, chr_rb.values), axis=0))
    chrom_df.columns = ['chrom', 'start', 'sample', 'svclass', 'size_bin', "length"]
    if chrom_df.shape[0] >= 10:
        # print(chrom_df['chrom'].unique())
        #assert(chrom_df['chrom'].nunique() == 1)

        #sort on 2nd column which is start coordinate
        chrom_df = chrom_df.sort_values(chrom_df.columns[1]) #CHROM, START

        coords = list(chrom_df[chrom_df.columns[1]])
        svtype = list(chrom_df.svclass)

        chrom_inter_distances = []

        #defined as the number of base pairs from one rearrangement breakpoint to the one immediately preceding it that is not its partner
        for i in range(1, len(coords)):
            j=i-1
            while j >= 0 and coords[j] == d[coords[i]]: #check if previous breakpoint is partner of this breakpoint, if it is, avoid it
                j=j-1
            dist = coords[i] - coords[j]
            chrom_inter_distances.append(dist)

        #now we take care of the edge cases of the first and last breakpoint
        if coords[1] == d[coords[0]]:
            first_dist = coords[2] - coords[0]
        else:
            first_dist = coords[1] - coords[0]

        chrom_inter_distances = [coords[0]] + chrom_inter_distances
        chrom_df['IMD'] = chrom_inter_distances

#     #INTERLEAVED VS NESTED CONFIGURATION
#     configuration = ['interleaved' for i in range(len(coords))]
#     for i in range(1, len(coords)):
#         j = i-1
#         while coords[j] == d[coords[i]] and not (d[coords[i]] < max(d[coords[j]], coords[j]) and coords[i] < max(d[coords[j]], coords[j]) and d[coords[i]] > min(d[coords[j]], coords[j]) and coords[i] > min(d[coords[j]], coords[j])): #check if previous breakpoint is partner of this breakpoint, if it is, avoid it
#             j=j-1
#         if j >= 0: #determine if we have a nested or interleaved configuration
#             if d[coords[i]] < max(d[coords[j]], coords[j]) and coords[i] < max(d[coords[j]], coords[j]) and d[coords[i]] > min(d[coords[j]], coords[j]) and coords[i] > min(d[coords[j]], coords[j]):
#                 configuration[i] = "nested"

#     chrom_df["Configuration"] = configuration
    return chrom_df

#major difference is whether its the closest breakpoint or the breakpoint immediately preceding it
#distance in bp to nearest breakpoint that is not it's partner (not distance to breakpoint immediately preceding)
def computeIMD3(chrom_df, chromosome):

    #keep track of partners

    d1 = dict(zip(list(chrom_df['start1']), list(chrom_df['start2'])))
    d2 = dict(zip(list(chrom_df['start2']), list(chrom_df['start1'])))
    d = {**d1, **d2} #combine dictionaries

    lb = chrom_df.iloc[:, 0:2] #get chrom1 and start1
    rb = chrom_df.iloc[:, 3:5] #get chrom2 and start2
    rest = chrom_df.iloc[:, 6:]

    lb = pd.DataFrame(np.concatenate((lb.values, rest.values), axis=1))
    rb = pd.DataFrame(np.concatenate((rb.values, rest.values), axis=1))

    #BREAKPOINTS ARE CONSIDERED INDIVIDUALLY

    #['chrom1', 'start1', 'end1', 'chrom2', 'start2', 'end2', 'sample', 'svclass', 'size_bin', 'length']
    lb.columns = ['chrom1', 'start1', 'sample', 'svclass', 'size_bin', "length"]
    rb.columns = ['chrom2', 'start2', 'sample', 'svclass', 'size_bin', "length"]

    chr_lb = lb[lb.chrom1 == chromosome]
    chr_rb = rb[rb.chrom2 == chromosome]
    # print(chr_lb)
    # print(chr_rb)
    chrom_df = pd.DataFrame(np.concatenate((chr_lb.values, chr_rb.values), axis=0))
    chrom_df.columns = ['chrom', 'start', 'sample', 'svclass', 'size_bin', "length"]

    # print(chrom_df['chrom'].unique())
    #assert(chrom_df['chrom'].nunique() == 1)

    #sort on last column which is start coordinate
    chrom_df = chrom_df.sort_values(chrom_df.columns[1]) #CHROM, START

    #take care of mirrored translocations
    to_drop = []
    starts = list(chrom_df["start"])
    svtypes = list(chrom_df["svclass"])
    for i, (s, svtype) in enumerate(zip(starts, svtypes)):
        if i+1 < len(starts) and abs(starts[i+1] - s) <= 100 and svtype == "translocation":
            to_drop.append(i)


    chrom_df = chrom_df.drop(to_drop)
    chrom_df = chrom_df.sort_values(chrom_df.columns[1])

    coords = list(chrom_df[chrom_df.columns[1]])
    svtype = list(chrom_df.svclass)

    chrom_inter_distances = []

    #defined as the number of base pairs from one rearrangement breakpoint to the one closest to it that is not it's partner
    for i in range(1, len(coords)-1):

        j = i-1
        k = i+1
        while j >= 0 and coords[j] == d[coords[i]]: #check if previous breakpoint is partner of this breakpoint, if it is, avoid it
            j=j-1
        while k < len(coords) and coords[k] == d[coords[i]]:
            k=k+1
        if j >= 0 and k < len(coords):
            if coords[i] - coords[j] == 0:
                dist = coords[k] - coords[i]
            elif coords[k] - coords[i] == 0:
                dist = coords[i] - coords[j]
            else:
                dist = min(coords[i] - coords[j], coords[k] - coords[i])
        elif j < 0:
            dist = coords[k] - coords[i]
        else:
            dist = coords[i] - coords[j]

        if dist == 0 and svtype[i] == "translocation":
            print(coords[j], coords[i], coords[k], dist)
            #print(len(coords))
        chrom_inter_distances.append(dist)
        if dist == 1:
            print(coords[j], coords[i], coords[k], svtype[i])

    #now we take care of the edge cases of the first and last breakpoint

    if coords[1] == d[coords[0]]:
        first_dist = coords[2] - coords[0]
    else:
        first_dist = coords[1] - coords[0]

    if coords[-2] == d[coords[-1]]:
        last_dist = coords[-1] - coords[-3]
    else:
        last_dist = coords[-1] - coords[-2]

    chrom_inter_distances = [first_dist] + chrom_inter_distances
    chrom_inter_distances.append(last_dist)
    chrom_df['IMD'] = chrom_inter_distances

    #INTERLEAVED VS NESTED CONFIGURATION
    configuration = ['interleaved' for i in range(len(coords))]
    for i in range(1, len(coords)):
        j = i-1
        while coords[j] == d[coords[i]] and not (d[coords[i]] < max(d[coords[j]], coords[j]) and coords[i] < max(d[coords[j]], coords[j]) and d[coords[i]] > min(d[coords[j]], coords[j]) and coords[i] > min(d[coords[j]], coords[j])): #check if previous breakpoint is partner of this breakpoint, if it is, avoid it
            j=j-1
        if j >= 0: #determine if we have a nested or interleaved configuration
            if d[coords[i]] < max(d[coords[j]], coords[j]) and coords[i] < max(d[coords[j]], coords[j]) and d[coords[i]] > min(d[coords[j]], coords[j]) and coords[i] > min(d[coords[j]], coords[j]):
                configuration[i] = "nested"

    chrom_df["Configuration"] = configuration
    return chrom_df

def unique_py(seqlist):
    seen = set()
    seen_add = seen.add
    return [x for x in seqlist if not (x in seen or seen_add(x))]

def calcIntermutDist(subs_type,first_chrom_na=False):
    subs_type_processed = pd.DataFrame()
    for c in unique_py(subs_type['chr']):
        subs_type_chrom = subs_type[subs_type['chr']==c].sort_values('position')
        if first_chrom_na:
            subs_type_chrom['prevPos'] = np.hstack((np.NAN,subs_type_chrom['position'].values.flatten()[:-1]))
        else:
            subs_type_chrom['prevPos'] = np.hstack((0,subs_type_chrom['position'].values.flatten()[:-1]))
        subs_type_chrom['distPrev'] = subs_type_chrom['position'].values - subs_type_chrom['prevPos'].values
        subs_type_processed = subs_type_processed.append(subs_type_chrom)
        subs_type_processed['distPrev'][subs_type_processed['distPrev']==0] = 1
    return subs_type_processed

def calcIntermutDist2(subs_type,first_chrom_na=False):
    subs_type_processed = subs_type.copy()
    chr_list = unique_py(subs_type['chr'])
    pos_array_im = subs_type['position'].values
    index_orig_df = np.arange(len(subs_type_processed))
    #args_pos_list = np.argsort(pos_array_im)
    args_pos_list=[]
    distPrev_list =[]
    prevPos_list =[]

    for c in chr_list:
        inds_chr = np.where(subs_type['chr']==c)
        pos_array_im_c = np.sort(pos_array_im[inds_chr])
        index_orig_df[inds_chr] = index_orig_df[inds_chr][np.argsort(pos_array_im[inds_chr])]

        if first_chrom_na:
            prevPos_arr_c =  np.hstack((np.NAN,pos_array_im_c.flatten()[:-1]))
        else:
            prevPos_arr_c = np.hstack((0,pos_array_im_c.flatten()[:-1]))
        distPrev_arr_c = pos_array_im_c  - prevPos_arr_c
        distPrev_arr_c[distPrev_arr_c==0] = 1
        distPrev_list = np.append(distPrev_list,distPrev_arr_c.astype(int)).flatten()
        prevPos_list = np.append(prevPos_list,prevPos_arr_c.astype(int)).flatten()
        prevPos_arr_c = []
        distPrev_arr_c = []
    subs_type_processed = subs_type_processed.reindex(index_orig_df).reset_index(drop=True)
    subs_type_processed['prevPos'] = prevPos_list
    subs_type_processed['distPrev'] = distPrev_list
    return subs_type_processed


def computeMAD(v):
    mad = np.median(np.abs(v-np.median(v)))
    return mad

def getMad(x,k=25):
    #Remove observations that are equal to zero; are likely to be imputed, should not contribute to sd:
    x = x[x!=0]
    runMedian = scisig.medfilt(x,k)
    dif = x-runMedian
    #SD = stats.median_abs_deviation(dif)
    SD = computeMAD(dif)
    return(SD)


def exactPcf(y, kmin, gamma, flag= True):
    if flag:
        yest = np.random.rand(len(y))
    else:
        yest = flag
    N    = len(y)
    yhat = np.zeros(N)
    if (N< 2*kmin):
        if (flag):
            results = {'Lengde':N,'sta':1,'mean':np.mean(y),'nIntervals':1,'yhat':np.repeat(np.mean(y),N, axis=0)}
            return results
        else:
            results = {'Lengde':N,'sta':1,'mean':np.mean(y),'nIntervals':1}
            return results

    initSum = sum(y[0:kmin])
    initKvad = sum(y[0:kmin]**2)
    initAve  = initSum/kmin
    bestCost = np.zeros(N)
    bestCost[kmin-1] = initKvad - initSum*initAve
    bestSplit = np.zeros(N)
    bestAver  = np.zeros(N)
    bestAver[kmin-1] = initAve
    Sum = np.zeros(N)
    Kvad = np.zeros(N)
    Aver = np.zeros(N)
    Cost = np.zeros(N)
    kminP1=kmin+1
    for k in range(kminP1,2*kmin):
        Sum[kminP1-1:k] = Sum[kminP1-1:k]+y[k-1]
        Aver[kminP1-1:k] = Sum[kminP1-1:k]/(range((k-kmin),0,-1))
        Kvad[kminP1-1:k] = Kvad[kminP1-1:k]+ (y[k-1]**2)
        bestAver[k-1] = (initSum+Sum[kminP1-1])/k
        bestCost[k-1] = (initKvad+Kvad[kminP1-1])-(k*bestAver[k-1]**2)

    for n in range(2*kmin,N+1):
        yn  = y[n-1]
        yn2 = y[n-1] **2
        Sum[kminP1-1:n] = Sum[kminP1-1:n]+yn
        Aver[kminP1-1:n] = Sum[kminP1-1:n]/(range((n-kmin),0,-1))
        Kvad[kminP1-1:n] = Kvad[kminP1-1:n]+yn2
        nMkminP1=n-kmin+1
        Cost[kminP1-1:nMkminP1] = bestCost[kmin-1:(n-kmin)]+Kvad[kminP1-1:nMkminP1]-Sum[kminP1-1:nMkminP1]*Aver[kminP1-1:nMkminP1]+gamma
        Pos = np.argmin(Cost[kminP1-1:nMkminP1])+kmin
        cost = Cost[Pos]
        aver = Aver[Pos]
        totAver = (Sum[kminP1-1]+initSum)/n
        totCost = (Kvad[kminP1-1]+initKvad) - n*totAver*totAver
        # if len(totCost)==0 or len(cost)==0 :
        #     raise ValueError('Something is Wrong')
        if totCost < cost :
            Pos = 1
            cost = totCost
            aver = totAver
        bestCost[n-1] = cost
        bestAver[n-1] = aver
        bestSplit[n-1] = Pos
    n=N
    antInt =1
    yest = np.array(yest,dtype=bool)
    bestSplit= np.array(bestSplit,dtype=int)

    if yest.any():
        while(n>0):
            yhat[(bestSplit[n-1]):n]  = bestAver[n-1]
            n = bestSplit[n-1]
            antInt = antInt+1
    else:
        while(n>0):
            n = bestSplit[n-1]
            antInt = antInt+1

    antInt =antInt -1
    #"""
    n = N
    lengde = np.repeat(0,antInt, axis=0)
    start = np.repeat(0,antInt, axis=0)
    verdi = np.repeat(0,antInt, axis=0)
    oldSplit = n
    antall = antInt
    while (n > 0):
        start[antall-1] = bestSplit[n-1]+1
        lengde[antall-1] = oldSplit-bestSplit[n-1]
        verdi[antall-1] = bestAver[n-1]
        n = bestSplit[n-1]
        oldSplit = n
        antall = antall-1
    if yest.any():
        results = {'Lengde':lengde,'sta':start,'mean':verdi,'nIntervals':antInt,'yhat':yhat}
        return results
    else:
        results = {'Lengde':lengde,'sta':start,'mean':verdi,'nIntervals':antInt}
        return results

def unique_py(seqlist):
    seen = set()
    seen_add = seen.add
    return [x for x in seqlist if not (x in seen or seen_add(x))]

def pbinom(q,size,prob=0.5):
    """
    Calculates the cumulative of the binomial distribution
    """
    result=binom.cdf(k=q,n=size,p=prob,loc=0)
    return result

def assignPvalues(kat_regions, chrom_bps, bp_rate = np.nan):
    if (len(kat_regions)>0):
        if np.isnan(bp_rate):
            left_bp  = min(chrom_bps['pos'])
            right_bp = max(chrom_bps['pos'])
            bp_rate = len(chrom_bps.values)/(right_bp - left_bp)
        kat_regions['pvalue'] = 1 - pbinom(kat_regions['number_bps'].values,kat_regions['end_bp'].values-kat_regions['start_bp'].values,bp_rate)
        kat_regions['d_seg'] = kat_regions['number_bps'].values / (kat_regions['end_bp'].values-kat_regions['start_bp'].values)
        kat_regions['rate_factor'] = kat_regions['d_seg'] / bp_rate
    return kat_regions

def assignPvalues2(kat_regions, chrom_bps, bp_rate = np.nan):
    if (len(kat_regions)>0):
        if np.isnan(bp_rate):
            bp_vals = chrom_bps['pos'].values
            left_bp  = np.min(bp_vals)
            right_bp = np.max(bp_vals)
            bp_rate = len(bp_vals)/(right_bp - left_bp)
        kat_regions['pvalue'] = 1 - pbinom(kat_regions['number_bps'].values,kat_regions['end_bp'].values-kat_regions['start_bp'].values,bp_rate)
        kat_regions['d_seg'] = kat_regions['number_bps'].values / (kat_regions['end_bp'].values-kat_regions['start_bp'].values)
        kat_regions['rate_factor'] = kat_regions['d_seg'] / bp_rate
    return kat_regions

def hotspotInfo(kat_regions_all,subs,segInterDist):
    if len(kat_regions_all) > 0 :
        kat_regions_all = kat_regions_all.reset_index(drop=True)
        for index in range(len(kat_regions_all)):
            subs_hotspot = subs[int(kat_regions_all['firstBp'][index]):int(kat_regions_all['lastBp'][index])+1]
            kat_regions_all['start_bp'][index] = min(subs_hotspot['pos'])
            kat_regions_all['end_bp'][index] = max(subs_hotspot['pos'])
            kat_regions_all['length_bp'][index] = kat_regions_all['end_bp'][index]-kat_regions_all['start_bp'][index]
            kat_regions_all['number_bps'][index] = len(subs_hotspot)
            if 'is_clustered' in subs_hotspot:
                kat_regions_all['number_bps_clustered'][index] = sum(subs_hotspot['is_clustered'])
            else:
                kat_regions_all['number_bps_clustered'][index] = 0
            if len(segInterDist) >0 & np.isnan(kat_regions_all['avgDist_bp'][index]):
                kat_regions_all['avgDist_bp'][index] = np.mean(segInterDist[int(kat_regions_all['firstBp'][index]):(int(kat_regions_all['lastBp'][index])+1)])
            kat_regions_all['no_samples'][index] = len(unique_py(list(subs_hotspot['sample'])))
            if 'pf' in subs_hotspot:
                kat_regions_all['no_del'][index] = len(subs_hotspot[subs_hotspot['pf']==2])
                kat_regions_all['no_dup'][index] = len(subs_hotspot[subs_hotspot['pf']==4])
                kat_regions_all['no_inv'][index] = len(subs_hotspot[subs_hotspot['pf']==1 | subs_hotspot['pf']==8])
                kat_regions_all['no_trn'][index] = len(subs_hotspot[subs_hotspot['pf']==32])
    return kat_regions_all

def hotspotInfo2(kat_regions_all,subs,segInterDist):
    if len(kat_regions_all) > 0 :
        pos_arr = subs['pos'].values
        kat_firstBp = kat_regions_all['firstBp'].values
        kat_lastBp = kat_regions_all['lastBp'].values
        kat_start_bp = kat_regions_all['start_bp'].values
        kat_end_bp = kat_regions_all['end_bp'].values
        kat_samples = list(subs['sample'])
        kat_regions_all = kat_regions_all.reset_index(drop=True)
        for index in range(len(kat_regions_all)):
            subs_hotspot = pos_arr[int(kat_firstBp[index]):int(kat_lastBp[index])+1]
            kat_regions_all['start_bp'][index] = np.min(subs_hotspot)
            kat_regions_all['end_bp'][index] = np.max(subs_hotspot)
            kat_regions_all['length_bp'][index] = kat_end_bp[index]-kat_start_bp[index]
            kat_regions_all['number_bps'][index] = len(subs_hotspot)
            if 'is_clustered' in kat_regions_all:
                subs_is_clust = kat_regions_all['is_clustered'].values[int(kat_firstBp[index]):int(kat_lastBp[index])+1]
                kat_regions_all['number_bps_clustered'][index] = np.sum(subs_is_clust)
            else:
                kat_regions_all['number_bps_clustered'][index] = 0
            if len(segInterDist) >0 & np.isnan(kat_regions_all['avgDist_bp'][index]):
                kat_regions_all['avgDist_bp'][index] = np.mean(segInterDist[int(kat_firstBp[index]):(int(kat_lastBp[index])+1)])
            kat_regions_all['no_samples'][index] = len(unique_py([kat_samples[val] for val in range(int(kat_firstBp[index]),int(kat_lastBp[index])+1)]))
            if 'pf' in kat_regions_all:
                kat_regions_all['no_del'][index] = len(subs_hotspot[subs_hotspot['pf']==2])
                kat_regions_all['no_dup'][index] = len(subs_hotspot[subs_hotspot['pf']==4])
                kat_regions_all['no_inv'][index] = len(subs_hotspot[subs_hotspot['pf']==1 | subs_hotspot['pf']==8])
                kat_regions_all['no_trn'][index] = len(subs_hotspot[subs_hotspot['pf']==32])
    return kat_regions_all

def extract_kat_regions(res,imd,subs,kmin_samples,pvalue_thresh,rate_factor_thresh,doMerging,kmin_filter,bp_rate):
    segInterDist = res['yhat']
    kataegis_threshold = imd
    kat_regions_all = pd.DataFrame()
    positions = subs['pos']
    katLoci = segInterDist <= kataegis_threshold # flag specifying if a point is in a peak
    if sum(katLoci>0):
        start_regions= np.asarray(np.where(katLoci[1:] & ~(katLoci[:-1]) | ((katLoci[1:] & (katLoci[:-1])) & (segInterDist[1:] != segInterDist[:len(katLoci)-1])) ))[0]+1
        if katLoci[0]:
            start_regions = np.hstack( (0 ,start_regions))
        end_regions =  np.asarray(np.where(~katLoci[1:] & (katLoci[:-1]) | ((katLoci[1:] & (katLoci[:-1])) & (segInterDist[1:] != segInterDist[:-1])) ))[0]
        if katLoci[-1]:
            end_regions = np.hstack( (end_regions,len(katLoci)-1))

        #handling Special cases
        if (len(end_regions)+len(start_regions)>0): # if there are any discontinuities in the segmentation at all
            if (len(end_regions)==1) & (len(start_regions)==0):
                start_regions = 0
            elif  (len(end_regions)==0) & (len(start_regions)==1):
                end_regions = len(positions)-1
            elif ((end_regions[0] < start_regions[0]) & (start_regions[-1]> end_regions[-1])):
                start_regions = np.hstack( (0 ,start_regions))
                end_regions = np.hstack( (end_regions,len(positions)-1))
            elif (end_regions[0] < start_regions[0]):
                # starts will be one shorter
                start_regions = np.hstack( (0 ,start_regions))
            elif (start_regions[-1] > end_regions[-1]):
                end_regions = np.hstack( (end_regions,len(positions)-1))
        # prepare a data structure that will be later filled up

        columnslist=['chr','start_bp','end_bp','length_bp','number_bps','number_bps_clustered','avgDist_bp','no_samples','no_del','no_dup','no_inv','np_trn','firstBp','lastBp']
        temp = matlib.repmat(np.nan,len(start_regions),len(columnslist))

        kat_regions_all = pd.DataFrame(temp,columns=columnslist)
        kat_regions_all['chr'] = subs['chr'][subs['chr'].index[0]]
        kat_regions_all['firstBp'] = start_regions
        kat_regions_all['lastBp']  = end_regions
        #print('intermittent= ', time.time()-t)
        #pdb.set_trace()
        #t = time.time()
        kat_regions_all= hotspotInfo2(kat_regions_all,subs,segInterDist)
        #print('hotspot1= ', time.time()-t)
        #pdb.set_trace()

        step_segInterDist_left = [np.nan] * len(segInterDist)
        step_segInterDist_left[1: len(segInterDist)] = segInterDist[1:len(segInterDist)] - segInterDist[0:len(segInterDist)-1]
        step_segInterDist_right = [np.nan] * len(segInterDist)
        step_segInterDist_right[0: len(segInterDist)-1] = segInterDist[0: len(segInterDist)-1] - segInterDist[1:len(segInterDist)]
        kat_regions_all['step_left'] = list(step_segInterDist_left[i] for i in start_regions)
        kat_regions_all['step_right'] = list(step_segInterDist_right[i] for i in end_regions)

        # run the filters on the regions of increased frequency
        # make sure there are at least kmin samples
        #t =time.time()
        if (not kat_regions_all.empty) & (len(kat_regions_all)>0) :
            kat_regions_all = kat_regions_all[kat_regions_all['no_samples'] >= kmin_samples]

        # make sure there are at least kmin.filter breakpoints
        if (not np.isnan(kmin_filter)):
            kat_regions_all = kat_regions_all[kat_regions_all['number_bps'] >= kmin_filter]

        if (not kat_regions_all.empty) & (len(kat_regions_all)>0) :
            kat_regions_all =   assignPvalues(kat_regions_all,subs,bp_rate)
            kat_regions_all = kat_regions_all[kat_regions_all['pvalue']<= pvalue_thresh]
            kat_regions_all = kat_regions_all[kat_regions_all['rate_factor']>= rate_factor_thresh]
        # merge segments if both were found to be peaks
        kat_regions_all=kat_regions_all.reset_index(drop=True)
        if (doMerging):
            if len(kat_regions_all)>1:
                for r in range(1,len(kat_regions_all)):
                    if kat_regions_all['lastBp'][r-1] ==  kat_regions_all['firstBp'][r] -1 :
                        # merge two segments
                        kat_regions_all['firstBp'][r] = kat_regions_all['firstBp'][r-1]
                        kat_regions_all['firstBp'][r-1] = np.nan
                        kat_regions_all['lastBp'][r-1] = np.nan
                        kat_regions_all['avgDist_bp'][r] = np.nan # this will need to be updated as segments are being merged
            # remove some of the merged segments
            columns_backup =  kat_regions_all.columns.to_list()
            kat_regions_all = kat_regions_all[ list(not( np.isnan(kat_regions_all['firstBp'].values[i])) and not np.isnan(kat_regions_all['lastBp'].values[i]) for i in range(len(kat_regions_all)))]
            if kat_regions_all.empty:
                kat_regions_all = pd.DataFrame(columns=columns_backup)
            kat_regions_all = hotspotInfo2(kat_regions_all, subs, segInterDist)
            kat_regions_all = assignPvalues(kat_regions_all, subs, bp_rate)
    return kat_regions_all

#######################################################

def annotateBedpe(sv_bedpe):
    #,kmin,kmin_samples,gamma_sdev=25,PEAK_FACTOR,thresh_dist,gamma,kmin_filter
    #sv_bedpe = data
    sv_bedpe['id'] = sv_bedpe.index + 1 #add an id to the rearrangement
    sv_bedpe = sv_bedpe.astype({"chrom1": str},{"chrom2": str})
    #functions below expect rows to be organised by chromosomes and ordered by position on the chromosome
    #prepare a dataframe for the calculation
    left  = pd.DataFrame(sv_bedpe[["chrom1","start1","sample","id"]])
    right = pd.DataFrame(sv_bedpe[["chrom2","start2","sample","id"]])
    left  = left.astype({"chrom1": str})
    right = right.astype({"chrom2": str})
    cncd  = pd.DataFrame(np.concatenate([left.values, right.values]), columns=('chr', 'position', 'sample','id'))
    cncd['isLeft']= True
    cncd['isLeft'][len(left):len(left)+len(right)]= False
    cncd = cncd[['chr', 'position', 'sample','isLeft','id']]

    sample_bps = pd.DataFrame(columns = cncd.columns)
    for chromi in unique_py(cncd['chr']):
        sample_bps= sample_bps.append(cncd[cncd['chr']==chromi].sort_values('position',kind='mergesort'),ignore_index = True)

    sample_bps.index = pd.RangeIndex(len(sample_bps.index))+1
    genome_size = 3 * 10**9
    MIN_BPS = 10   # minimal number of breakpoints on a chromosome to do any any segmentation
    logScale = False
    exp_dist = genome_size/len(sample_bps)
    gamma_sdev  = 25 #
    PEAK_FACTOR = 10
    thresh_dist = np.NaN

    if logScale:
        sample_bps['intermut_dist'] = np.log10(calcIntermutDist2(sample_bps,first_chrom_na=False)['distPrev'].values)
        if np.isnan(thresh_dist):
            thresh_dist = np.log10(exp_dist/PEAK_FACTOR)
    else:
        sample_bps['intermut_dist'] = calcIntermutDist2(sample_bps,first_chrom_na=False)['distPrev'].values
        if np.isnan(thresh_dist):
            thresh_dist = exp_dist/PEAK_FACTOR

    gamma = np.NaN
    if np.isnan(gamma) & ~np.isnan(gamma_sdev):
        # compute the mean absolute deviation
        sdev = getMad(sample_bps['intermut_dist'].values)
        gamma = gamma_sdev*sdev

    sample_bps['is_clustered_single'] = False
    all_kat_regions = pd.DataFrame()
    sample_bps['mean_intermut_dist']= np.NaN
    for chrom in unique_py(sample_bps['chr']): # loop over chromosomes
        sample_bps_flag = sample_bps['chr'] == chrom  #breakpoints on a current chromosome
        if sum(sample_bps_flag) > MIN_BPS:  # if there are enough breakpoints on a chromosome to run pcf
            data_points = sample_bps['intermut_dist'][sample_bps_flag]
            kmin =10
            res = exactPcf(data_points.values, kmin, gamma, True)
            sample_bps['mean_intermut_dist'][sample_bps_flag] = res['yhat']
            # prepare the points for pcf
            subs = pd.DataFrame(columns=['chr','pos','sample'])
            subs['chr'] = sample_bps['chr'][sample_bps_flag]
            subs['pos'] = sample_bps['position'][sample_bps_flag]
            subs['sample'] = sample_bps['sample'][sample_bps_flag]
            kmin_samples, kmin_filter, doMerging, pvalue_thresh, rate_factor_thresh,bp_rate = 1, kmin, True, 1,1,np.nan
            kat_regions = extract_kat_regions(res,thresh_dist,subs,kmin_samples,pvalue_thresh,rate_factor_thresh,doMerging, kmin_filter,bp_rate)

            all_kat_regions = pd.concat([all_kat_regions, kat_regions], axis =0)
            if not kat_regions.empty & len(kat_regions)>0 :
                for k in range(len(kat_regions)):
                    ind = np.where(sample_bps_flag)[0]
                    temp =sample_bps['is_clustered_single'].values[ind]
                    temp[int(kat_regions['firstBp'][k]):int(kat_regions['lastBp'][k])+1]=True
                    sample_bps['is_clustered_single'][ind[temp]+1] = True
        else:
            sample_bps['mean_intermut_dist'][sample_bps_flag]= np.mean(sample_bps['intermut_dist'][sample_bps_flag])

    if not logScale: # even if pcf was run on non-logged distances, the output is logged
        sample_bps['intermut_dist'] = np.log10(sample_bps['intermut_dist'].values.astype(float))
        sample_bps['mean_intermut_dist'] = np.log10(sample_bps['mean_intermut_dist'].values.astype(float))
    # a rearrangement is in a cluster if any of its breakpoints are

    sample_bps['is_clustered'] = sample_bps['is_clustered_single']
    sv_bedpe['is_clustered'] = np.nan

    check_exist_list = sample_bps['id'][sample_bps['is_clustered']]
    sample_bps['is_clustered'][np.in1d(sample_bps['id'].values,check_exist_list.values)] = True
    sv_bedpe['is_clustered'] = np.in1d(sv_bedpe['id'],sample_bps['id'][sample_bps['is_clustered']])
    sv_bedpe = processBEDPE(sv_bedpe)
    result={'sv_bedpe':sv_bedpe,'kat_regions':all_kat_regions}
    return result

def generateSVMatrix(input_dir, project, output_dir, skip=False):
    # create output_dir if it does not yet exist
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    if input_dir[-1] != "/":
        input_dir = input_dir + "/"
    all_samples = [] #list of dataframes for each sample
    for f in os.listdir(input_dir):
        if os.path.isfile(input_dir + f):
            print("Generating count vector for " + f)
            data = pd.read_csv(input_dir + f,sep='\t')
            if data.shape[0] == 0:
                print("SKIPPING " + str(f) + "because it has 0 SVs")
                continue
            elif ("sample" not in data.columns or len(data["sample"].iloc[0]) <= 1 or "chrom1" not in data.columns or "chrom2" not in data.columns or "start1" not in data.columns or "start2" not in data.columns or "end1" not in data.columns or "end2" not in data.columns) and skip==False:
                raise Exception("Please ensure that there is a sample column containing the name of the sample")
            elif ("sample" not in data.columns or len(data["sample"].iloc[0]) <= 1 or "chrom1" not in data.columns or "chrom2" not in data.columns or "start1" not in data.columns or "start2" not in data.columns or "end1" not in data.columns or "end2" not in data.columns) and skip==True:
                print("Warning: it appears that " + str(f) + " may not have the correct input format, please check for required columns that are missing")
                continue
            else:
                #get annotated bedpe for a single sample
                result= annotateBedpe(data)

            all_samples.append(result['sv_bedpe'])
    matrix = tsv2matrix(all_samples, project, output_dir)
    out_file = os.path.join(output_dir, project + ".SV32.matrix.tsv")
    matrix.to_csv(out_file, sep="\t")
    print("Saved matrix to " + out_file)
    plotSV(matrix, output_dir, project, plot_type="pdf", percentage=False, aggregate=True)
    plot_file_name = os.path.join(output_dir, project + '_RS32_counts_aggregated.pdf')
    print("Saved aggregate SV32 plot to " + plot_file_name)
    return matrix


#reformat input bedpe files
def processBEDPE(df):
    """A function that processes a given bedpe file produced by an SV caller"""



    #CHECK FORMAT OF CHROMOSOME COLUMN ("chr1" vs. "1"), needs to be the latter
    if not str(df['chrom1'][0]).isdigit():
        if df['chrom1'][0].startswith("chr"):
            chrom1 = []
            chrom2 = []
            for a, b in zip(df['chrom1'], df['chrom2']):
                if a.startswith("chr") or b.startswith("chr"):
                    a = a[3:]
                    b = b[3:]
                    chrom1.append(a)
                    chrom2.append(b)
                else:
                    break

            df['chrom1'] = chrom1
            df['chrom2'] = chrom2

    #df = df[(df["chrom1"] != 'Y') & (df["chrom2"] != 'Y')]


    if "strand1" in df.columns and "strand2" in df.columns:
        df = df[["chrom1", "start1", "end1", "chrom2", "start2", "end2", "strand1", "strand2", "sample", "is_clustered"]]
    else:
        df = df[["chrom1", "start1", "end1", "chrom2", "start2", "end2", "sample", "svclass", "is_clustered"]]
    df = df.astype({df.columns[1]: 'int32', df.columns[2]: 'int32', df.columns[4]: 'int32', df.columns[5]: 'int32', df.columns[0]: 'str', df.columns[3]: 'str'})

    lengths = []
    if "svclass" not in df.columns:
        if "strand1" not in df.columns or "strand2" not in df.columns:
            raise Exception("cannot classify rearrangements: svclass column missing, and cannot compute it because strand1 and strand2 are missing.")
        else:
            svclass = []
            for row in df.itertuples():
                if row.chrom1 != row.chrom2:
                    sv = "translocation"
                    svclass.append(sv)
                    #print(row)
                elif (row.strand1 == '+' and row.strand2 == '-') or (row.strand1 == '-' and row.strand2 == '+'):
                    sv = "inversion"
                    svclass.append(sv)
                elif row.strand1 == '+' and row.strand2 == '+':
                    sv = "deletion"
                    svclass.append(sv)
                elif row.strand1 == '-' and row.strand2 == '-':
                    sv = "tandem-duplication"
                    svclass.append(sv)
                else:
                    raise Exception("cannot classify rearrangements: svclass column missing, and cannot compute it because strand1 and strand2 are not in the proper format.")
            #f.write(svclass)
            df["svclass"] = svclass
    else:
        svclass = list(df["svclass"])

    #GET SIZE
    sizes = [0 for x in svclass]
    i=-1
    for row in df.itertuples():
        i=i+1
        if row.svclass != "translocation":
            lengths.append(abs(row.start1 - row.start2))
            l = abs(row.start1 - row.start2) / 1000000 #covert to megabases
            # if abs(row.start1 - row.start2) < 1000:
            #     print(row.svclass, abs(row.start1 - row.start2), row.sample, row.is_clustered)
            if l <= 0.010:
                size = "1-10Kb"
                sizes[i] = size
            elif l > 0.01 and l <= 0.1:
                size = "10-100Kb"
                sizes[i] = size
            elif l > 0.1 and l <= 1:
                size = "100Kb-1Mb"
                sizes[i] = size
            elif l > 1 and l <= 10:
                size = "1Mb-10Mb"
                sizes[i] = size
            else:
                size = ">10Mb"
                sizes[i] = size
        else:
            sizes[i] = "0"
            lengths.append(abs(row.start1 - row.start2))
            #print(row)

    df["size_bin"] = sizes
    df["length"] = lengths

    df = df.filter(items=['chrom1', 'start1', 'end1', 'chrom2', 'start2', 'end2', 'sample', 'svclass', 'size_bin', 'length', "is_clustered"])

    to_remove = []
    #remove SV's less than 1KB (unless its a translocation)
    for row in df.itertuples():
        index = row.Index
        if row.svclass != "translocation" and row.length < 1000:
            to_remove.append(index)
            #print(row)

    df.drop(df.index[to_remove], inplace=True)
    return df

def tsv2matrix(sv_bedpe_list, project, output_dir):
    features = ['clustered_del_1-10Kb', 'clustered_del_10-100Kb', 'clustered_del_100Kb-1Mb', 'clustered_del_1Mb-10Mb', 'clustered_del_>10Mb', 'clustered_tds_1-10Kb', 'clustered_tds_10-100Kb', 'clustered_tds_100Kb-1Mb', 'clustered_tds_1Mb-10Mb', 'clustered_tds_>10Mb', 'clustered_inv_1-10Kb', 'clustered_inv_10-100Kb', 'clustered_inv_100Kb-1Mb', 'clustered_inv_1Mb-10Mb', 'clustered_inv_>10Mb', 'clustered_trans', 'non-clustered_del_1-10Kb', 'non-clustered_del_10-100Kb', 'non-clustered_del_100Kb-1Mb', 'non-clustered_del_1Mb-10Mb', 'non-clustered_del_>10Mb', 'non-clustered_tds_1-10Kb', 'non-clustered_tds_10-100Kb', 'non-clustered_tds_100Kb-1Mb', 'non-clustered_tds_1Mb-10Mb', 'non-clustered_tds_>10Mb', 'non-clustered_inv_1-10Kb', 'non-clustered_inv_10-100Kb', 'non-clustered_inv_100Kb-1Mb', 'non-clustered_inv_1Mb-10Mb', 'non-clustered_inv_>10Mb', 'non-clustered_trans']
    svclass_mapping = {"deletion":"del", "tandem-duplication":"tds", "inversion":"inv", "translocation":"trans"}
    if len(sv_bedpe_list) <= 1:
        warnings.warn("There seems to be <= 1 samples, please ensure the sample column contains a unique sample name")
    df = pd.concat(sv_bedpe_list) #one master table with all samples
    out_file = os.path.join(output_dir, project + ".SV32.annotated.tsv")
    df.to_csv(out_file, index=False, sep="\t")
    print("Saved annotated bedpe to " + out_file)
    samples = list(df["sample"].unique())
    arr = np.zeros((32, len(samples)), dtype='int')
    nmf_matrix = pd.DataFrame(arr, index=features, columns=samples)
    for row in df.itertuples():
        if row.is_clustered:
            c = "clustered"
        else:
            c = "non-clustered"

        if svclass_mapping[row.svclass] != "trans":
            channel = c + "_" + svclass_mapping[row.svclass] + "_" + row.size_bin
        else:
            channel = c + "_" + svclass_mapping[row.svclass]
        nmf_matrix.at[channel, row.sample] += 1
    nmf_matrix.reindex([features])
    nmf_matrix.index.name = 'MutationType'
    #nmf_matrix.reindex([features]).reset_index()

    return nmf_matrix

def plotSV(matrix, output_path, project, plot_type="pdf", percentage=False, aggregate=False):
    #inner function to construct plot
    def plot(counts, labels, sample, project, percentage, aggregate=False):

        if percentage:
            counts = [(x/sum(counts))*100 for x in counts]

        color_mapping = {'del':{'>10Mb':"deeppink", '1Mb-10Mb':"hotpink", '10-100Kb':"lightpink", '100Kb-1Mb':"palevioletred", '1-10Kb':"lavenderblush"},
                     'tds':{'>10Mb':"saddlebrown", '1Mb-10Mb':"sienna", '10-100Kb':"sandybrown", '100Kb-1Mb':"peru", '1-10Kb':"linen"},
                     'inv':{'>10Mb':"rebeccapurple", '1Mb-10Mb':"blueviolet", '10-100Kb':"plum", '100Kb-1Mb':"mediumorchid", '1-10Kb':"thistle"}}

        alpha_dict = dict(enumerate(string.ascii_lowercase))
        x_labels = ['1-10kb', '10-100kb', '100kb-1Mb', '1Mb-10Mb','>10Mb']
        super_class = ['clustered', 'non-clustered']
        sub_class = ['del', 'tds', 'inv', 'trans']
        N=32
        ticks = [0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31]
        width = 0.27
        xticks = []
        i = -1 #used to distinguish first bar from the rest
        fig, ax = plt.subplots(figsize=(16,8))

        # Custom Formatting
        plt.style.use('ggplot')
        plt.rcParams['axes.facecolor'] = 'white'
        plt.gca().yaxis.grid(True)
        plt.gca().grid(which='major', axis='y', color=[0.93,0.93,0.93], zorder=1)
        ax.set_axisbelow(True)
        ax.yaxis.set_major_locator(ticker.LinearLocator(5))
        ax.spines["bottom"].set_color("black")
        ax.spines["top"].set_color("black")
        ax.spines["right"].set_color("black")
        ax.spines["left"].set_color("black")
        plt.xlim(xmin=-.5,xmax=len(labels)-.5)
        tmp_max=max(counts)
        plt.ylim(ymax=1.25*tmp_max)
        # Add light gray horizontal lines at y-ticks
        ax.grid(linestyle='-', linewidth=1, color='#EDEDED', axis='y')

        for count, label in zip(counts, labels):
            categories = label.split('_')
            if len(categories) > 2:
                rearrangement_class = categories[1]
                size_class = categories[2]
            i += 1 #position of bar
            #f.write (categories)

            if len(categories) == 2: #clustered translocation or non-clustered translocation
                ax.bar(ticks[i], count, color="dimgray", edgecolor='black') #translocation only has one color
            else:
                ax.bar(ticks[i], count, color=color_mapping[rearrangement_class][size_class], edgecolor='black')

            xticks.append(ticks[i])
        ax.set_xticks(xticks)
        ax.set_xticklabels(x_labels * 3 + [' '] + x_labels * 3 + [' '], rotation=90, weight="bold", fontsize = 16, fontname='Arial', color='black')
        ax.tick_params(labelleft=True, left=False, bottom=False)
        ax.tick_params(axis='y', which='major', pad=0, labelsize=30)

        #ADD PATCHES AND TEXT
        patch_height = 0.05
        patch_width = 2.8
        loh_width= 2.5
        loh_len = 4.8

        trans = transforms.blended_transform_factory(ax.transData, ax.transAxes)

        #### CLUSTERED PATCHES ####
        ax.add_patch(plt.Rectangle((-.5, 1.095), 15.9, patch_height*1.5, clip_on=False, facecolor='gray', transform=trans))
        plt.text(6, 1.1125, "Clustered", fontsize=23, fontname='Arial', fontweight='bold', color='white', transform=trans)
        ax.add_patch(plt.Rectangle((-.5, 1.01), loh_len+.1, patch_height*1.5, clip_on=False, facecolor='maroon', transform=trans))
        plt.text(1.3, 1.03, "Del", fontsize=23, fontname='Arial', fontweight='bold', color='white', transform=trans)
        ax.add_patch(plt.Rectangle((4.6, 1.01), loh_len, patch_height*1.5, clip_on=False, facecolor='darkorange', transform=trans))
        plt.text(6.27, 1.03, "Tds", fontsize=23, fontname='Arial', fontweight='bold', color='white', transform=trans)
        ax.add_patch(plt.Rectangle((9.6, 1.01), loh_len, patch_height*1.5, clip_on=False, facecolor='slateblue', transform=trans))
        plt.text(11.35, 1.03, "Inv", fontsize=23, fontname='Arial', fontweight='bold', color='white', transform=trans)
        ax.add_patch(plt.Rectangle((14.6, 1.01), .8, patch_height*1.5, clip_on=False, facecolor='dimgray', transform=trans))
        plt.text(14.75, 1.03, "T", fontsize=23, fontname='Arial', fontweight='bold', color='white', transform=trans)

        # add vertical black lines
        ax.axvline(x=15.5, color='black', linewidth=1)

        #### NON-CLUSTERED PATCHES ####
        ax.add_patch(plt.Rectangle((15.6, 1.095), 15.9, patch_height*1.5, clip_on=False, facecolor='black', transform=trans))
        plt.text(21, 1.1125, "Non-Clustered", fontsize=23, fontname='Arial', fontweight='bold', color='white', transform=trans)
        ax.add_patch(plt.Rectangle((15.6, 1.01), loh_len, patch_height*1.5, clip_on=False, facecolor='maroon', transform=trans))
        plt.text(17.35, 1.03, "Del", fontsize=23, fontname='Arial', fontweight='bold', color='white', transform=trans)
        ax.add_patch(plt.Rectangle((20.6, 1.01), loh_len, patch_height*1.5, clip_on=False, facecolor='darkorange', transform=trans))
        plt.text(22.25, 1.03, "Tds", fontsize=23, fontname='Arial', fontweight='bold', color='white', transform=trans)
        ax.add_patch(plt.Rectangle((25.6, 1.01), loh_len, patch_height*1.5, clip_on=False, facecolor='slateblue', transform=trans))
        plt.text(27.37, 1.03, "Inv", fontsize=23, fontname='Arial', fontweight='bold', color='white', transform=trans)
        ax.add_patch(plt.Rectangle((30.6, 1.01), .9, patch_height*1.5, clip_on=False, facecolor='dimgray', transform=trans))
        plt.text(30.82, 1.03, "T", fontsize=23, fontname='Arial', fontweight='bold', color='white', transform=trans)


        # format the set_yticklabels labels
        if percentage:
            tmp_y_labels =['{0:0.1f}%'.format(round(x,1)) for x in ax.get_yticks().tolist()]
        else:
            tmp_y_labels =[round(x,1) for x in ax.get_yticks().tolist()]
        #ax.yaxis.labelpad = 300

        # set the y-axis labels
        ax.set_yticklabels(tmp_y_labels, fontname='Arial', weight='bold', fontsize=16, color='black')

        #y-axis titles
        if aggregate:
            ax.set_ylabel("# of events per sample", fontsize=24, fontname="Arial", weight = 'bold', labelpad = 15, color='black')
        elif percentage:
            ax.set_ylabel("Percentage(%)", fontsize=24, fontname="Arial", weight = 'bold', labelpad = 15, color='black')
            #ax.yaxis.labelpad = 1
        else:
            ax.set_ylabel("# of events", fontsize=24, fontname="Arial", weight = 'bold', labelpad = 15, color='black')

        #TITLE
        if not aggregate:
            plt.text(0, 0.90, sample, fontsize=20, fontname='Arial', fontweight='bold', color='black', transform=trans)
        else:
            plt.text(0, 0.90, project, fontsize=20, fontname='Arial', fontweight='bold', color='black', transform=trans)

        plt.show()
        pp.savefig(fig, dpi=600, bbox_inches='tight')
    df = matrix.copy()
    labels = list(df.index)
    #labels = list(df[label])
    if aggregate:
        num_samples = len(df.columns) - 1
        df['total_count'] = df.sum(axis=1) / num_samples #NORMALIZE BY # of SAMPLES
        counts = list(df['total_count'])
        sample = ''
        output_plot_path = os.path.join(output_path, project + '_RS32_counts_aggregated.pdf')
        pp = PdfPages(output_plot_path)
        plot(counts, labels, sample, project, percentage, aggregate=True)
    else:
        if plot_type == 'pdf' and percentage:
            output_plot_path = os.path.join(output_path, project + '_RS32_signatures.pdf')
            pp = PdfPages(output_plot_path)
        elif plot_type == 'pdf' and percentage==False:
            output_plot_path = os.path.join(output_path, project + '_RS32_counts.pdf')
            pp = PdfPages(output_path + project + '_RS32_counts.pdf')
        else: #input == counts
            f.write("The only plot type supported at this time is pdf")

        #each column vector in dataframe contains counts for a specific sample
        samples = list(df)[1:]
        for i, (col, sample) in enumerate(zip(df.columns[1:], samples)):
            counts = list(df[col])
            counts = [(x/sum(counts))*100 for x in counts]
            assert(len(counts)) == 32
            plot(counts, labels, sample, project, percentage)
    pp.close()

if __name__ == "__main__":
        if len(sys.argv) > 1:
            input_dir, project, output_dir = sys.argv[1], sys.argv[2], sys.argv[3]
        generateSVMatrix(input_dir, project, output_dir)
